import json
import subprocess
from threading import Thread
import numpy as np
import pandas as pd
import cv2
import PIL
import requests
import uuid
import os
import sys
import traceback
import inspect
import re
import base64
import yaml
import zlib
import hashlib
import select

import shutil
import tempfile
from typing import List, Tuple, Dict, Optional

from naeural_core.utils.thread_raise import ctype_async_raise

try:
  # Temporarily guard the bs4 import until we can be sure
  # that it's available in all environments.
  import bs4
except ImportError as _:
  bs4 = None


from collections import OrderedDict, defaultdict, deque
from io import BufferedReader, BytesIO
from time import sleep, time
from datetime import datetime, timedelta, timezone
from copy import deepcopy
from xml.etree import ElementTree
from urllib.parse import urlparse, urlunparse
from functools import partial

from naeural_core import constants as ct

from naeural_core.serving.ai_engines.utils import (
  get_serving_process_given_ai_engine,
  get_ai_engine_given_serving_process,
  get_params_given_ai_engine
)

from naeural_core.utils.plugins_base.persistence_serialization_mixin import _PersistenceSerializationMixin
from naeural_core.utils.system_shared_memory import NumpySharedMemory

from naeural_core.main.ver import __VER__ as core_version    
from ratio1._ver import __VER__ as sdk_version   

COMMON_COLORS = {
    "red_1":     (255, 0, 0),
    "green_1":   (0, 128, 0),
    "green_2":   (0, 50, 0),
    "blue_1":    (0, 0, 255),
    "yellow_1":  (255, 255, 0),

    "yellow_2"  :  (255, 215, 0),
    "orange_1":  (255, 165, 0),
    "purple_1" : (128, 0, 128),
    "blue_2":    (0, 200, 200),  # Bright bluish-green
    "blue_3":    (20, 90, 150),    # Dark blue
    "green_3"  :  (0, 128, 128),
    "brown_1":   (128, 0, 0),    # brownish-red
    "brown_2" :  (139, 69, 19), # chocolate brown
    "pink_1":    (220, 185, 190), # A soft pink; tweak as desired
    "grey_1":  (192, 192, 192),  # A lighter grey variant
    "white_1":   (255, 255, 255),
    "grey_2":    (50, 50, 50),
    "black_1":   (0, 0, 0)
}


GIT_IGNORE_AUTH = ["-c","http.https://github.com/.extraheader="]

class NestedDotDict(dict):
  # TODO: maybe use https://github.com/mewwts/addict/blob/master/addict/addict.py
  __getattr__ = defaultdict.__getitem__
  __setattr__ = defaultdict.__setitem__
  __delattr__ = defaultdict.__delitem__

  def __init__(self, *args, **kwargs):
    super(NestedDotDict, self).__init__(*args, **kwargs)
    for key, value in self.items():
      if isinstance(value, dict):
        self[key] = NestedDotDict(value)
      elif isinstance(value, (list, tuple)):
        self[key] = type(value)(
          NestedDotDict(v) if isinstance(v, dict) else v for v in value
        )        
              
  def __deepcopy__(self, memo):
    return NestedDotDict({k: deepcopy(v, memo) for k, v in self.items()})
                
  def __reduce__(self):
    return (self.__class__, (), self.__getstate__())

  def __getstate__(self, obj=None):
    state = {}
    obj = obj or self
    for key, value in obj.items():
      if isinstance(value, NestedDotDict):
        state[key] = self.__getstate__(value)
      else:
        state[key] = value
    return state

  def __setstate__(self, state):
    self.update(state)
  
class DefaultDotDict(defaultdict):
  __getattr__ = defaultdict.__getitem__
  __setattr__ = defaultdict.__setitem__
  __delattr__ = defaultdict.__delitem__
  

class NestedDefaultDotDict(defaultdict):
  """
  A dictionary-like object supporting auto-creation of nested dictionaries and default values for undefined keys.
  """
  def __init__(self, *args, **kwargs):
    super(NestedDefaultDotDict, self).__init__(NestedDefaultDotDict, *args, **kwargs)
    for key, value in dict(self).items():
      if isinstance(value, dict):
        self[key] = NestedDefaultDotDict(value)
      elif isinstance(value, (list, tuple)):
        self[key] = type(value)(
          NestedDefaultDotDict(v) if isinstance(v, dict) else v for v in value
        )

  def __getattr__(self, item):
    if item in self:
      return self[item]
    return self.__missing__(item)

  def __setattr__(self, key, value):
    if isinstance(value, dict) and not isinstance(value, NestedDefaultDotDict):
      value = NestedDefaultDotDict(value)
    defaultdict.__setitem__(self, key, value)

  def __delattr__(self, item):
    try:
      defaultdict.__delitem__(self, item)
    except KeyError as e:
      raise AttributeError(e)

  def __deepcopy__(self, memo):
    return NestedDefaultDotDict({k: deepcopy(v, memo) for k, v in self.items()})

  def __reduce__(self):
    return (self.__class__, (), None, None, iter(self.items()))


class NPJson(json.JSONEncoder):
  """
  Used to help jsonify numpy arrays or lists that contain numpy data types.
  """
  def default(self, obj):
      if isinstance(obj, np.integer):
          return int(obj)
      elif isinstance(obj, np.floating):
          return float(obj)
      elif isinstance(obj, np.ndarray):
          return obj.tolist()
      elif isinstance(obj, np.ndarray):
          return obj.tolist()
      elif isinstance(obj, datetime):
          return obj.strftime("%Y-%m-%d %H:%M:%S")
      else:
          return super(NPJson, self).default(obj)


class LogReader():
  def __init__(self, owner, buff_reader, size=100):
    self.buff_reader: BufferedReader = buff_reader
    self.owner = owner
    self.buf_reader_size = size

    self.__fd = buff_reader.fileno()
    self.__fd_reading = False

    self.buffer = []
    self.done = False
    self.exited = False
    self.thread = None
    # now we start the thread
    self.start()
    return

  def P(self, msg, color=None, **kwargs):
    """
    Print a message to the owner process.
    """
    self.owner.P(msg, color=color, **kwargs)
    return

  def _do_fd_read(self):
    if not self.__fd_reading:
      self.__fd_reading = True
      os.set_blocking(self.__fd, False)
    chunk = None
    try:
      chunk = os.read(self.__fd, self.buf_reader_size)
    except:
      pass
    return chunk

  def _do_select_read(self):
    ready, _, _ = select.select([self.buff_reader], [], [], 0.1)  # Wait up to 0.1s
    text = None
    if ready:
      text = self.buff_reader.read(self.buf_reader_size)
    return text

  def maybe_close_buffer(self):
    if not self.buff_reader.closed:
      self.P("Closing buff reader...")
      self.buff_reader.close()
      self.P("Buff reader closed.")
    # endif buff_reader not closed yet
    return

  def _run(self):
    force_stop = False
    log_msg = None
    try:
      while not self.done:
        # text = self._do_fd_read()
        text = self._do_select_read()
        if text:  # Check if any data is read
          self.on_text(text)
        else:
          pass # break can lead to early exit so nothing to do
        # endif any data ready
      # self.exited needs to be set to True as soon as the loop is exited
      # in order to avoid any forced exception during the finally block.
      log_msg = "Log reader loop finished."
    except ct.ForceStopException as exc:
      force_stop = True
    except Exception as exc:
      log_msg = f"Log reader exception: {exc}"
    self.exited = True
    if not force_stop:
      try:
        if log_msg:
          self.P(log_msg)
        self.maybe_close_buffer()
        self.P("Log reader stopped.")
      except Exception as e:
        pass
    # endif force_stop
    return

  def on_text(self, text):
    # if isinstance(text, bytes):
    #   # Decode bytes to string
    #   text = text.decode('utf-8', errors='replace')
    self.buffer.append(text)
    return

  def start(self):
    self.thread = Thread(
      target=self._run,
      daemon=True,
    )
    self.thread.start()
    return

  # Public methods
  def stop(self):
    if self.done:
      return
    self.done = True
    self.P("Stopping log reader thread...")

    if not self.exited:
      seconds = 5
      self.P(f"Waiting {seconds} for log reader thread to stop...")
      self.owner.sleep(seconds)
    # end if

    if not self.exited:
      self.P("Forcing log reader thread to stop...")
      self.maybe_close_buffer()

      with self.owner.log.managed_lock_logger():
        ctype_async_raise(self.thread.ident, ct.ForceStopException)
      self.owner.sleep(0.2)
      self.P("Log reader stopped forcefully.")
    # end if

    self.P("Joining log reader thread...")
    self.thread.join(timeout=0.1)
    self.P("Log reader thread joined.")

    if self.thread.is_alive():
      self.P("Log reader thread is still alive.", color='r')
    else:
      self.P("Log reader thread joined gracefully.")
    # end if

    return

  def get_next_characters(self, max_characters=-1, decode='utf-8', decode_errors='replace'):
    result = []
    
    if max_characters == -1:
      # get all the buffer
      L = len(self.buffer)
      for i in range(L):
        result.append(self.buffer.pop(0))
      # end for
    else:
      L = len(self.buffer)
      nr_chars = 0
      for i in range(L):
        segment = self.buffer[0]

        if nr_chars + len(segment) > max_characters:
          result.append(segment[:max_characters - nr_chars])
          self.buffer[0] = self.buffer[0][max_characters - nr_chars:]
          break
        elif nr_chars + len(segment) == max_characters:
          result.append(segment)
          self.buffer.pop(0)
          break
        else:
          result.append(segment)
          nr_chars += len(segment)
          self.buffer.pop(0)
        # end if
      # end for
    # end if
    result = b''.join(result)
    if decode is None:
      return result
    if decode == False:
      return result
    if decode == True:
      return result.decode('utf-8', errors=decode_errors)
    if len(decode) > 0:
      return result.decode(decode, errors=decode_errors)
    return result

  def get_next_line(self):
    result = []
    L = len(self.buffer)
    for i in range(L):
      segment = self.buffer[i]
      if '\n' in segment:
        pos = segment.index('\n')
        result.append(segment[:pos])
        self.buffer[i] = segment[pos+1:]
        break
      else:
        result.append(segment)
      # end if
    # end for

    if len(result) > 0 and '\n' in result[-1]:
      for _ in range(len(result)):
        self.buffer.pop(0)
      result = ''.join(result)
    else:
      result = None
    # end if

    return result


class _UtilsBaseMixin(
  _PersistenceSerializationMixin
  ):

  def __init__(self):
    super(_UtilsBaseMixin, self).__init__()
    return
  
  
  @property
  def ee_core_ver(self):
    return core_version
  
  @property
  def ee_sdk_ver(self):
    return sdk_version

  
  def trace_info(self):
    """
    Returns a multi-line string with the last exception stacktrace (if any)

    Returns
    -------
    str.

    """
    return traceback.format_exc()
  
  
  def python_version(self):
    """
    Utilitary method for accessing the Python version.
    Returns
    -------
    Version of python
    """
    return sys.version.split()[0]
  
  def get_serving_process_given_ai_engine(self, ai_engine):
    return get_serving_process_given_ai_engine(ai_engine)
  
    

  def timedelta(self, **kwargs):
    """
    Alias of `datetime.timedelta`
    

    Parameters
    ----------
    **kwargs : 
      can contain days, seconds, microseconds, milliseconds, minutes, hours, weeks.


    Returns
    -------
    timedelta object


    Example
    -------
      ```
        diff = self.timedelta(seconds=10)
      ```
    
    """
    return timedelta(**kwargs)  
  
  
  def time(self):
    """
    Returns current timestamp

    Returns
    -------
    time : timestamp (float)
      current timestamp.
      
      
    Example
    -------
      ```
      t1 = self.time()
      ... # do some stuff
      elapsed = self.time() - t1
      ```    

    """

    return time() 

  def now_str(self, nice_print=False, short=False):
    """
    Returns current timestamp in string format
    Parameters
    ----------
    nice_print
    short

    Returns
    -------

    """
    return self.log.now_str(nice_print=nice_print, short=short)

  def get_output_folder(self):
    """
    Provides access to get_output_folder() method from .log
    Returns
    -------

    """
    return self.log.get_output_folder()

  def get_data_folder(self):
    """
    Provides access to get_data_folder() method from .log
    Returns
    -------

    """
    return self.log.get_data_folder()

  def get_logs_folder(self):
    """
    Provides access to get_logs_folder() method from .log
    Returns
    -------

    """
    return self.log.get_logs_folder()

  def get_models_folder(self):
    """
    Provides access to get_models_folder() method from .log
    Returns
    -------

    """
    return self.log.get_models_folder()

  def get_target_folder(self, target):
    """
    Provides access to get_target_folder() method from .log
    Parameters
    ----------
    target

    Returns
    -------

    """
    return self.log.get_target_folder(target)

  def sleep(self, seconds):
    """
    sleeps current job a number of seconds
    """
    sleep(seconds)
    return  


  def uuid(self, size=13):
    """
    Returns a unique id.
  

    Parameters
    ----------
    size : int, optional
      the number of chars in the uid. The default is 13.

    Returns
    -------
    str
      the uid.
      

    Example
    -------
    
      ```
        str_uid = self.uuid()
        result = {'generated' : str_uid}
      ```      

    """
    return str(uuid.uuid4())[:size].replace('-','')
  
  @property
  def json(self):
    """
    Provides access to `json` package

    Returns
    -------
    `json` package      

    """
    return json

  @property
  def yaml(self):
    """
    Provides access to `yaml` package

    Returns
    -------
    `yaml` package      

    """
    return yaml

  @property
  def re(self):
    """
    Provides access to `re` package

    Returns
    -------
    `re` package

    """
    return re
  
  @property
  def inspect(self):
    """
    Provides access to `inspect` package

    Returns
    -------
    `inspect` package      

    """
    return inspect
    
  
  @property
  def requests(self):
    """
    Provides access to `requests` package

    Returns
    -------
    `requests` package      

    """
    return requests

  @property
  def urlparse(self):
    """
    Provides access to `urlparse` method from `urllib.parse` package

    Returns
    -------
    `urlparse` method      

    """
    return urlparse

  @property
  def urlunparse(self):
    """
    Provides access to `urlunparse` method from `urllib.parse` package

    Returns
    -------
    `urlunparse` method      

    """
    return urlunparse
  
  @property
  def consts(self):
    """
    Provides access to E2 constants

    Returns
    -------
    ct : package
      Use `self.consts.CONST_ACME` to acces any required constant

    """
    return ct


  @property
  def const(self):
    """
    Provides access to E2 constants

    Returns
    -------
    ct : package
      Use `self.const.ct.CONST_ACME` to acces any required constant

    """
    return ct

  @property
  def ct(self):
    """
    Provides access to E2 constants

    Returns
    -------
    ct : package
      Use `self.const.ct.CONST_ACME` to acces any required constant

    """
    return ct

  @property
  def ds_consts(self):
    """
    Alias for DatasetBuilder class from E2 constants
    Provides access to constants used in DatasetBuilderMixin
    Returns
    -------
    ct.DatasetBuilder : package
      Use `self.ds_consts.CONST_ACME` to access any required constant
    """
    return ct.DatasetBuilder

  @property
  def cv2(self):
    """
    provides access to computer vision library
    """
    return cv2

  @property
  def np(self):
    """
    Provides access to numerical processing library
    

    Returns
    -------
    np : Numpy package
      
    Example:
      ```
      np_zeros = self.np.zeros(shape=(10,10))
      ```
    """
    return np  
  
  @property
  def OrderedDict(self):
    """
    Returns the definition for `OrderedDict`

    Returns
    -------
    OrderedDict : class
      `OrderedDict` from standard python `collections` package.
      
    Example
    -------
        ```
        dct_A = self.OrderedDict({'a': 1})
        dct_A['b'] = 2
        ```

    """
    return OrderedDict  
  
  
  @property
  def defaultdict(self):
    """
    provides access to defaultdict class


    Returns
    -------
      defaultdict : class
      
    Example
    -------
      ```
        dct_integers = self.defaultdict(lambda: 0)
      ```

    """
    return defaultdict
  
  
  def DefaultDotDict(self, *args):
    """
    Returns a `DefaultDotDict` object that is a `dict` where you can use keys with dot 
    using the default initialization
    
    Inputs
    ------
    
      pass a `lambda: <type>` always
    
    Returns
    -------
      DefaultDotDict : class
     
    Example
    -------
     ```
       dct_dot = self.DefaultDotDict(lambda: str)
       dct_dot.test1 = "test"       
       print(dct_dot.test1)
       print(dct_dot.test2)
     ```

    """
    return DefaultDotDict(*args)
  
  def NestedDotDict(self, *args):
    """
    Returns a `NestedDotDict` object that is a `dict` where you can use keys with dot
    
    Returns
    -------
      defaultdict : class
     
    Example
    -------
     ```
       dct_dot = self.NestedDotDict({'test' : {'a' : 100}})
       dct_dot.test.a = "test"   
       print(dct_dot.test.a)
    """
    return NestedDotDict(*args)
  
  
  def NestedDefaultDotDict(self, *args):
    """
    Returns a `NestedDefaultDotDict` object that is a `defaultdict(dict)` where you can use keys with dot
    
    Returns
    -------
      defaultdict : class
     
    Example
    -------
     ```
      dct_dot1 = self.NestedDefaultDotDict()
      dct_dot1.test.a = "test"   
      print(dct_dot1.test.a)
       
      dct_dot2 = self.NestedDefaultDotDict({'test' : {'a' : 100, 'b' : {'c' : 200}}})
      print(dct_dot2.test.a)
      print(dct_dot2.test.b.c)
      print(dct_dot2.test.b.unk)
        
    """
    return NestedDefaultDotDict(*args)

  def LogReader(self, buff_reader, size=100):
    """
    Returns a `LogReader` object that is used to read from a buffer reader.

    Parameters
    ----------
    buff_reader : BufferedReader
        the buffer from where to read
    size : int, optional
        the size of the buffer. The default is 100.

    Returns
    -------
    LogReader : class
        the log reader object.
    """

    return LogReader(owner=self, buff_reader=buff_reader, size=size)

  def path_exists(self, path):
    """
    TODO: must be reviewed
    """
    return self.os_path.exists(path)
  
  
  @property
  def deque(self):
    """
    provides access to deque class
    """
    return deque  
  
  @property
  def datetime(self):
    """
    Proxy for the `datetime.datetime`

    Returns
    -------
      datetime : datetime object
      
      
    Example
    -------
      ```
      now = self.datetime.now()
      ```

    """
    return datetime

  @property
  def timezone(self):
    """
    Proxy for the `datetime.timezone`

    Returns
    -------
      timezone : timezone object
      
      
    Example
    -------
      ```
      utc = self.timezone.utc
      ```

    """
    return timezone

  @property
  def deepcopy(self):
    """
    This method allows us to use the method deepcopy
    """
    return deepcopy
  
  @property
  def os_path(self):
    """
    Proxy for `os.path` package


    Returns
    -------
      package
      
      
    Example
    -------
      ```
      fn = self.diskapi_save_dataframe_to_data(df, 'test.csv')
      exists = self.os_path.exists(fn)
      ```

    """
    return os.path
  
  @property
  def os_environ(self):
    """
    Returns a copy of the current environment variables based on `os.environ`.
    Important: Changing a value in the returned dictionary does NOT change 
               the value of the actual environment variable.
    

    Returns
    -------
    _type_
        _description_
    """
    return os.environ.copy()

  @property
  def PIL(self):
    """
    provides access to PIL package
    """
    return PIL

  @property
  def BytesIO(self):
    """
    provides access to BytesIO class from io package
    """
    return BytesIO

  @property
  def ElementTree(self):
    """
    provides access to ElementTree class from xml.etree package
    """
    return ElementTree

  @property
  def pd(self):
    """
    Provides access to pandas library

    Returns
    -------
      package
      
      
    Example
    -------
      ```
      df = self.pd.DataFrame({'a' : [1,2,3], 'b':[0,0,1]})      
      ```

    """
    return pd  

  @property
  def partial(self):
    """
    Provides access to `functools.partial` method

    Returns
    -------
      method


    Example
    -------
      ```
      fn = self.partial(self.diskapi_save_dataframe_to_data, fn='test.csv')
      ```

    """
    return partial

  def safe_json_dumps(self, dct, replace_nan=False, **kwargs):
    """Safe json dumps that can handle numpy arrays and so on

    Parameters
    ----------
    dct : dict
        The dict to be dumped
        
    replace_nan : bool, optional
        Replaces nan values with None. The default is False.

    Returns
    -------
    str
        The json string
    """
    return self.log.safe_json_dumps(dct, replace_nan=replace_nan, **kwargs)

  
  def json_dumps(self, dct, replace_nan=False, **kwargs):
    """Alias for `safe_json_dumps` for backward compatibility
    """
    return self.safe_json_dumps(dct, replace_nan=replace_nan, **kwargs)
  
  def json_loads(self, json_str, **kwargs):
    """
    Parses a json string and returns the dictionary
    """
    return self.json.loads(json_str, **kwargs)
  
  
  def load_config_file(self, fn):
    """
    Loads a json/yaml config file and returns the config dictionary

    Parameters
    ----------
    fn : str
      The filename of the config file

    Returns
    -------
    dict
      The config dictionary
    """
    return self.log.load_config_file(fn=fn)
  
  
  def maybe_download(self, url, fn, target='output', **kwargs):
    """
    Enables http/htps/minio download capabilities.


    Parameters
    ----------
    url : str or list
      The URI or URIs to be used for downloads
      
    fn: str of list
      The filename or filenames to be locally used
      
    target: str
      Can be `output`, `models` or `data`. Default is `output`

    kwargs: dict
      if url starts with 'minio:' the function will retrieve minio conn
             params from **kwargs and use minio_download (if needed or forced)

    Returns
    -------
      files, messages : list, list
        all the local files and result messages from download process
      
      
    Example
    -------
    """
    res = None
    files, msgs = self.log.maybe_download(
      url=url,
      fn=fn,
      target=target,
      **kwargs,
    )
    if len(files) >= 1:
      if len(files) == 1:
        res = files[0]
      else:
        res = files
    else:
      self.P('Errors while downloading: {}'.format([str(x) for x in msgs]))
    return res

  """GIT SECTION"""
  if True:
    def __utils_log(self, msg, color=None, **kwargs):
      """
      Helper method for fallback in case your class doesn't have a self.P method.
      Parameters
      ----------
      msg : str
      color : str, optional
      """
      if hasattr(self, 'P'):
        self.P(msg, color=color, **kwargs)
      else:
        print(msg)
      return

    def git_clone(self, repo_url, repo_dir, target='output', user=None, token=None, pull_if_exists=True):
      """
      Clones a git repository or pulls if the repository already exists.

      Parameters
      ----------
      repo_url : str
        The git repository URL

      token : str, optional
        The token to be used for authentication. The default is None.

      user: str, optional
        The username to be used for authentication. The default is None.

      token : str, optional
        The token to be used for authentication. The default is None.

      pull_if_exists : bool, optional
        If True, the repository will be pulled if it already exists. The default is True.


      Returns
      -------
      str
        The local folder where the repository was cloned.
      """

      repo_path = self.os_path.join(self.get_target_folder(target), repo_dir)
      self.__utils_log(f"git_clone: '{repo_url}' to '{repo_path}'")

      if user is not None and token is not None:
        repo_url = repo_url.replace('https://', f'https://{user}:{token}@')

      USE_GIT_IGNORE_AUTH = True # for git pull -c does not work

      try:
        command = None
        if self.os_path.exists(repo_path) and pull_if_exists:
          # Repository already exists, perform git pull
          self.__utils_log(f"git_clone: Repo exists at {repo_path} -> pulling...")
          if USE_GIT_IGNORE_AUTH:
            command = ["git"] + GIT_IGNORE_AUTH + ["pull"]
          else:
            command = ["git", "pull"]
          results = subprocess.check_output(
              command,
              cwd=repo_path,
              stderr=subprocess.STDOUT,
              universal_newlines=True,
              # creationflags=subprocess.CREATE_NO_WINDOW, # WARNING: This works only on Windows
          )
        else:
          # Clone the repository
          if USE_GIT_IGNORE_AUTH:
            command = ["git"] + GIT_IGNORE_AUTH + ["clone", repo_url, repo_path]
          else:
            command = ["git", "clone", repo_url, repo_path]
          results = subprocess.check_output(
              command,
              stderr=subprocess.STDOUT,
              universal_newlines=True,
              # creationflags=subprocess.CREATE_NO_WINDOW, # WARNING: This works only on Windows
          )
        # end if
        self.__utils_log(f"git_clone: `{' '.join(command)}` results:\n{results}")
      except subprocess.CalledProcessError as exc:
        self.__utils_log(f"git_clone: Error '{exc.cmd}' returned  {exc.returncode} with output:\n{exc.output}", color='r')
        repo_path = None
      except Exception as exc:
        self.__utils_log(f"git_clone: Error while cloning git repository {repo_url} in {repo_path}: {exc}", color='r  ')
        repo_path = None
      # end try
      return repo_path


    def git_checkout_tag(self, repo_dir, tag):
      cmd = ["git", "-C", repo_dir, "checkout", f"tags/{tag}"]
      subprocess.check_call(cmd)
      return


    def git_get_local_commit_hash(self, repo_dir):
      """
      Retrieves the latest commit hash from the local git repository.

      Parameters
      ----------
      repo_dir : str
        The local directory where the repository is cloned.

      Returns
      -------
      str
        The latest commit hash from the local repository.
      """
      commit_hash = None
      self.__utils_log(f"git_get_local_commit_hash: {repo_dir}")

      command = ["git", "rev-parse", "HEAD"]
      try:
        results = subprocess.check_output(
          command,
          cwd=repo_dir,
          stderr=subprocess.STDOUT,
          universal_newlines=True,
          # creationflags=subprocess.CREATE_NO_WINDOW, # WARNING: This works only on Windows
        )
        if results is not None:
          self.__utils_log(f"git_get_local_commit_hash: `rev-parse` results:\n{results}")
          lines = results.split('\n')
          if len(lines) > 0:
            commit_hash = lines[0].split()[0]
        else:
          self.__utils_log(
            f"git_get_local_commit_hash: Error while retrieving commit hash from remote repository: {results}",
            color='r'
          )
      except subprocess.CalledProcessError as exc:
        self.__utils_log(
          f"git_get_local_commit_hash: Error '{exc.cmd}' returned  {exc.returncode} with output:\n{exc.output}",
          color='r'
        )
      except Exception as exc:
        self.__utils_log(
          f"git_get_local_commit_hash: An unexpected exception occurred: {exc}",
          color='r'
        )
      return commit_hash


    def git_get_last_commit_hash(self, repo_url, user=None, token=None):
      """
      Retrieves the latest commit hash from the remote git repository.

      Parameters
      ----------
      repo_url : str
        The git repository URL

      user : str, optional
        The username to be used for authentication. The default is None.

      token : str, optional
        The token to be used for authentication. The default is None.

      Returns
      -------
      str
        The latest commit hash from the remote repository.
      """
      commit_hash = None
      self.__utils_log(f"git_get_last_commit_hash: using {repo_url}")

      if user is not None and token is not None:
        repo_url = repo_url.replace('https://', f'https://{user}:{token}@')

      command = ["git"] + GIT_IGNORE_AUTH + ["ls-remote", repo_url, "HEAD"]
      try:
        results = subprocess.check_output(
          command,
          stderr=subprocess.STDOUT,
          universal_newlines=True,
          # creationflags=subprocess.CREATE_NO_WINDOW, # WARNING: This works only on Windows
        )
        if results is not None:
          self.__utils_log(f"git_get_last_commit_hash: `ls-remote` results:\n{results}")
          lines = results.split('\n')
          if len(lines) > 0:
            commit_hash = lines[0].split()[0]
        else:
          self.__utils_log(
            f"git_get_last_commit_hash: Error while retrieving commit hash from remote repository: {results}",
            color='r'
          )
      except subprocess.CalledProcessError as exc:
        self.__utils_log(
          f"git_get_last_commit_hash: Error '{exc.cmd}' returned  {exc.returncode} with output:\n{exc.output}",
          color='r'
        )
      except Exception as exc:
        self.__utils_log(
          f"git_get_last_commit_hash: An unexpected exception occurred: {exc}",
          color='r'
        )
      return commit_hash


    def git_get_latest_release_asset_info(
        self,
        repo_url: str,
        token: str = None,
        release_tag_substring: str = None,
        asset_filter: str = None
    ):
      """
      1) Query GitHub releases for `repo_url`.
      2) Filter by `release_tag_substring`, pick newest by published_at.
      3) From that release's assets, pick the first containing `asset_filter`.
      4) Return a dict with { "release_tag": ..., "asset_id": ..., "asset_name": ...,
                              "asset_updated_at": ..., "published_at": ... }
      or None if not found.
      """
      # Parse https://github.com/owner/repo -> (owner, repo)
      segments = repo_url.rstrip("/").split("/")
      if len(segments) < 2:
        self.__utils_log(f"[ERROR] Invalid repo URL: {repo_url}", color='r')
        return None
      owner, repo = segments[-2], segments[-1]

      api_url = f"https://api.github.com/repos/{owner}/{repo}/releases"
      headers = {}
      if token:
        headers["Authorization"] = f"token {token}"

      # Fetch releases
      try:
        resp = requests.get(api_url, headers=headers, timeout=10)
        if resp.status_code != 200:
          self.__utils_log(f"[ERROR] Could not fetch releases: {resp.status_code} {resp.text}", color='r')
          return None
        releases = resp.json()
      except Exception as exc:
        self.__utils_log(f"[ERROR] Exception fetching releases: {exc}", color='r')
        return None

      if not isinstance(releases, list) or not releases:
        self.__utils_log("[INFO] No releases found or invalid data.", color='y')
        return None

      # Filter by release_tag_substring
      matching = []
      for rel in releases:
        tag_name = rel.get("tag_name", "")
        if release_tag_substring is None or (release_tag_substring in tag_name):
          matching.append(rel)
      if not matching:
        self.__utils_log(f"[INFO] No release matches '{release_tag_substring}'.", color='y')
        return None

      # Sort by published_at descending
      matching.sort(key=lambda r: r.get("published_at", ""), reverse=True)
      newest_release = matching[0]
      newest_tag = newest_release.get("tag_name", "")
      assets = newest_release.get("assets", [])

      if not assets:
        self.__utils_log("[INFO] This release has no assets.", color='y')
        return None

      # Find the first asset that matches asset_filter
      chosen_asset = None
      for asset in assets:
        if asset_filter is None or (asset_filter in asset["name"]):
          chosen_asset = asset
          break
      if not chosen_asset:
        self.__utils_log(f"[INFO] No asset matches filter '{asset_filter}'.", color='y')
        return None

      # Return relevant metadata
      info = {
        "release_tag": newest_tag,
        "asset_id": chosen_asset.get("id"),
        "asset_name": chosen_asset.get("name"),
        "asset_updated_at": chosen_asset.get("updated_at"),
        "published_at": newest_release.get("published_at"),
      }
      return info

    def _filter_and_pick_newest_by_substring(
        self, items: List[Tuple[str, int]],
        substring: Optional[str] = None
    ) -> Optional[str]:
      """
      Given a list of (name, date_as_int), filter by substring in `name`,
      then pick the one with the greatest `date_as_int`.

      Returns the `name` of the newest item, or None if no matches.
      """
      if not items:
        return None

      # Filter by substring
      filtered = []
      for (name, date_val) in items:
        if substring is None or (substring in name):
          filtered.append((name, date_val))

      if not filtered:
        return None

      # Sort by date_val ascending, pick the last (newest)
      filtered.sort(key=lambda x: x[1])
      return filtered[-1][0]

    def git_clone_and_list_tags(
        self,
        repo_url: str,
        user: Optional[str] = None,
        token: Optional[str] = None
    ) -> List[Tuple[str, int]]:
      """
      1) Clone/fetch tags from `repo_url` into a temp directory (using `git`).
      2) Return a list of (tag_name, creation_date_unix).
      """
      # Insert credentials if provided and if repo_url starts with https://
      if user and token and repo_url.startswith("https://"):
        repo_url = repo_url.replace("https://", f"https://{user}:{token}@", 1)

      self.__utils_log(f"Fetching tags from {repo_url}...")

      temp_dir = tempfile.mkdtemp(prefix="git_tags_")
      try:
        # Initialize empty repo
        subprocess.check_call(["git", "init", temp_dir],
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.DEVNULL)

        # Add remote origin
        subprocess.check_call(["git", "-C", temp_dir, "remote", "add", "origin", repo_url],
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.DEVNULL)

        # Fetch *all* tags (depth=1 typically suffices)
        subprocess.check_call([
          "git", "-C", temp_dir, "fetch", "--tags", "--depth=1", "origin"
        ], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

        # List tags sorted by creation date
        cmd = [
          "git", "-C", temp_dir, "for-each-ref",
          "--sort=creatordate",
          '--format=%(refname:short)___%(creatordate:unix)',
          "refs/tags"
        ]
        output = subprocess.check_output(cmd, universal_newlines=True)
        lines = output.strip().splitlines()
        if not lines:
          self.__utils_log("[WARNING] No tags found in the repository.")
          return []

        # Parse lines: "<tag>___<date_unix>"
        results = []
        for line in lines:
          parts = line.split("___")
          if len(parts) != 2:
            continue
          tag, date_unix_str = parts
          try:
            date_unix = int(date_unix_str)
          except ValueError:
            continue
          results.append((tag, date_unix))

        return results

      except subprocess.CalledProcessError as exc:
        self.__utils_log(f"[ERROR] Command {exc.cmd} failed with code={exc.returncode}")
        return []
      except Exception as exc:
        self.__utils_log(f"[ERROR] Unexpected error: {exc}")
        return []
      finally:
        # Cleanup
        shutil.rmtree(temp_dir, ignore_errors=True)
      return

    def git_get_last_release_tag(
        self,
        repo_url: str,
        user: Optional[str] = None,
        token: Optional[str] = None,
        tag_name: Optional[str] = None
    ) -> Optional[str]:
      """
      Return the newest (by creation date) tag from the given Git repo
      that contains `tag_name` as a substring (if provided).
      """
      tags = self.git_clone_and_list_tags(repo_url, user, token)
      if not tags:
        return None

      newest_tag = self._filter_and_pick_newest_by_substring(tags, tag_name)
      if newest_tag:
        self.__utils_log(f"Newest matching tag = {newest_tag}")
      else:
        self.__utils_log(f"[INFO] No tags match substring '{tag_name}'.")
      return newest_tag


    def git_download_release_asset(
        self,
        repo_url: str,
        user: Optional[str] = None,  # not used except for logging or if you want to unify
        token: Optional[str] = None,
        release_tag_substring: Optional[str] = None,
        asset_filter: Optional[str] = None,
        download_dir: Optional[str] = None
    ) -> Optional[str]:
      """
      1) Gets info about the most recent GitHub release asset for `repo_url`.
      2) Downloads it to `download_dir`.
      3) Returns the path to the downloaded file, or None on error.

      Also returns None if no matching release or asset is found.
      """
      # 1) First, get the newest release + asset
      info = self.git_get_latest_release_asset_info(
        repo_url, token=token,
        release_tag_substring=release_tag_substring,
        asset_filter=asset_filter
      )
      if not info:
        return None  # Something failed or no match

      owner, repo = repo_url.rstrip("/").split("/")[-2:]
      asset_id = info["asset_id"]
      file_name = info["asset_name"]

      if download_dir is None:
        download_dir = os.path.join(os.getcwd(), "github_assets")
      os.makedirs(download_dir, exist_ok=True)

      asset_api_url = f"https://api.github.com/repos/{owner}/{repo}/releases/assets/{asset_id}"
      local_path = os.path.join(download_dir, file_name)

      headers = {"Accept": "application/octet-stream"}
      if token:
        headers["Authorization"] = f"token {token}"

      self.__utils_log(f"Downloading asset '{file_name}' to '{local_path}'...")
      try:
        with requests.get(asset_api_url, headers=headers, stream=True, timeout=30) as r:
          r.raise_for_status()
          with open(local_path, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
              f.write(chunk)
      except Exception as exc:
        self.__utils_log(f"[ERROR] Download failed: {exc}", color='r')
        return None

      self.__utils_log(f"Asset downloaded to: {local_path}")
      return local_path
  """END GIT SECTION"""


  def indent_strings(self, strings, indent=2):
    """ Indents a string or a list of strings by a given number of spaces."""
    lst_strings = strings.split('\n')
    lst_strings = [f"{' ' * indent}{string}" for string in lst_strings]
    result = '\n'.join(lst_strings)
    return result
  


  def dict_to_str(self, dct:dict):
    """
    Transforms a dict into a pre-formatted strig without json package

    Parameters
    ----------
    dct : dict
      The given dict that will be string formatted.

    Returns
    -------
    str
      the nicely formatted.
      
      
    Example:
    -------
      ```
      dct = {
        'a' : {
          'a1' : [1,2,3]
        },
        'b' : 'abc'
      }
      
      str_nice_dict = self.dict_to_str(dct=dct)
      ```

    """
    return self.log.dict_pretty_format(dct)  
  
  def timestamp_to_str(self, ts=None, fmt='%Y-%m-%d %H:%M:%S'):
    """
    Returns the string representation of current time or of a given timestamp


    Parameters
    ----------
    ts : float, optional
      timestamp. The default is None and will generate string for current timestamp. 
    fmt : str, optional
      format. The default is '%Y-%m-%d %H:%M:%S'.


    Returns
    -------
    str
      the timestamp in string format.
      
    
    Example
    -------
        
      ```
      t1 = self.time()
      ...
      str_t1 = self.time_to_str(t1)
      result = {'T1' : str_t1}
      ```
    """
    if ts is None:
      ts = self.time()
    return self.log.time_to_str(t=ts, fmt=fmt)
  
  
  def time_to_str(self, ts=None, fmt='%Y-%m-%d %H:%M:%S'):
    """
    Alias for `timestamp_to_str`
    

    Parameters
    ----------
    ts : float, optional
      The given time. The default is None.
    fmt : str, optional
      The time format. The default is '%Y-%m-%d %H:%M:%S'.

    Returns
    -------
    str
      the string formatted time.
      
      
    Example
    -------
      ```
      t1 = self.time()
      ...
      str_t1 = self.time_to_str(t1)
      result = {'T1' : str_t1}
      ```

    """
    return self.timestamp_to_str(ts=ts, fmt=fmt)
  
  
  def datetime_to_str(self, dt=None, fmt='%Y-%m-%d %H:%M:%S'):
    """
    Returns the string representation of current datetime or of a given datetime

    Parameters
    ----------
    dt : datetime, optional
      a given datetime. The default is `None` and will generate string for current date.
    fmt : str, optional
      datetime format. The default is '%Y-%m-%d %H:%M:%S'.

    Returns
    -------
    str
      the datetime in string format.
      
    
    Example
    -------
      ```
      d1 = self.datetime()
      ...
      str_d1 = self.datetime_to_str(d1)
      result = {'D1' : str_d1}
      ```
    

    """
    if dt is None:
      dt = datetime.now()
    return datetime.strftime(dt, format=fmt)

  def time_in_interval_hours(self, ts, start, end):
    """
    Provides access to method `time_in_interval_hours` from .log
    Parameters
    ----------
    ts: datetime timestamp
    start = 'hh:mm'
    end = 'hh:mm'

    Returns
    -------

    """
    return self.log.time_in_interval_hours(ts, start, end)

  def time_in_schedule(self, ts, schedule, weekdays=None):
    """
    Check if a given timestamp `ts` is in a active schedule given the schedule data


    Parameters
    ----------
    ts : float
      the given timestamp.
      
    schedule : dict or list
      the schedule.
            
    weekdays : TYPE, optional
      list of weekdays. The default is None.


    Returns
    -------
    bool
      Returns true if time in schedule.
      

    Example
    -------
      ```
      simple_schedule = [["09:00", "12:00"], ["13:00", "17:00"]]
      is_working = self.time_in_schedule(self.time(), schedule=simple_schedule)
      ```

    """
    return self.log.time_in_schedule(
      ts=ts,
      schedule=schedule,
      weekdays=weekdays
    )
    
    
  


  def now_in_schedule(self, schedule, weekdays=None):
    """
    Check if the current time is in a active schedule given the schedule data


    Parameters
    ----------
    schedule : dict or list
      the schedule.
            
    weekdays : TYPE, optional
      list of weekdays. The default is None.


    Returns
    -------
    bool
      Returns true if time in schedule.
      

    Example
    -------
      ```
      simple_schedule = [["09:00", "12:00"], ["13:00", "17:00"]]
      is_working = self.now_in_schedule(schedule=simple_schedule)
      ```

    """
    return self.log.now_in_schedule(
      schedule=schedule,
      weekdays=weekdays
    )  
    
    
  def img_to_base64(self, img):
    """Transforms a numpy image into a base64 encoded image

    Parameters
    ----------
    img : np.ndarray
        the input image

    Returns
    -------
    str: base64 encoded image
    """
    return self.log.np_image_to_base64(img)

  def base64_to_img(self, b64):
    """
    Transforms a base64 encoded image into a np.ndarray
    Parameters
    ----------
    b64 : str
      the base64 image
    Returns
    -------
    np.ndarray: the decoded image
    """
    return self.log.base64_to_np_image(b64)

  
  def base64_to_str(self, b64, decompress=False, url_safe=False):
    """Transforms a base64 encoded string into a normal string

    Parameters
    ----------
    b64 : str
        the base64 encoded string
        
    decompress : bool, optional
        if True, the string will be decompressed after decoding. The default is False.

    Returns
    -------
    str: the decoded string
    """
    b_encoded = b64.encode('utf-8')
    if url_safe:
      b_text = base64.urlsafe_b64decode(b_encoded)
    else:
      b_text = base64.b64decode(b_encoded)
      
    if decompress:
      b_text = zlib.decompress(b_text)
    str_text = b_text.decode('utf-8')
    return str_text
  

  def base64_to_bytes(self, b64, decompress=False, url_safe=False) -> bytes:
    """Transforms a base64 encoded string into bytes

    Parameters
    ----------
    b64 : str
        the base64 encoded string
        
    decompress : bool, optional
        if True, the data will be decompressed after decoding. The default is False.

    Returns
    -------
    bytes: the decoded data
    """
    b_encoded = b64.encode('utf-8')
    if url_safe:
      b_text = base64.urlsafe_b64decode(b_encoded)
    else:
      b_text = base64.b64decode(b_encoded)
      
    if decompress:
      b_text = zlib.decompress(b_text)
    
    return b_text  
  

  def execute_remote_code(self, code: str, debug: bool = False, timeout: int = 10):
    """
    Execute code received remotely.
    Parameters
    ----------
    code : str
        the code to be executed
    debug : bool, optional
        if True, the code will be executed in debug mode. The default is False.
    timeout : int, optional
        the timeout for the code execution. The default is 10.
    Returns
    -------
    dict: the result of the code execution
    If the code execution was successful, the result will contain the following keys:
    - result: the result of the code execution
    - errors: the errors that occurred during the execution
    - warnings: the warnings that occurred during the execution
    - prints: the printed messages during the execution
    - timestamp: the timestamp of the execution
    If the code execution failed, the result will contain the following key:
    - error: the error message
    """
    if not isinstance(code, str):
      return {'error': 'Code must be a string'}
    if len(code) == 0:
      return {'error': 'Code should not be an empty string'}
    result, errors, warnings, printed = None, None, [], []
    self.P(f'Executing code:\n{code}')
    b64_code, errors = self.code_to_base64(code, return_errors=True)
    if errors is not None:
      return {'error': errors}
    res = self.exec_code(
      str_b64code=b64_code,
      debug=debug,
      self_var='plugin',
      modify=True,
      return_printed=True,
      timeout=timeout
    )
    if isinstance(res, tuple):
      result, errors, warnings, printed = res
    return {
      'result': result,
      'errors': errors,
      'warnings': warnings,
      'prints': printed,
      'timestamp': self.time()
    }


  def shorten_str(self, s, max_len=32):
    """
    Shortens a string to a given max length.
    Parameters
    ----------
    s : str | list | dict
    max_len : int, optional

    Returns
    -------
    str | list | dict : the shortened string
    """
    if isinstance(s, str):
      return s[:max_len] + '...' if len(s) > max_len else s
    if isinstance(s, list):
      return [self.shorten_str(x, max_len) for x in s]
    if isinstance(s, dict):
      return {k: self.shorten_str(v, max_len) for k, v in s.items()}
    return s

  def normalize_text(self, text):
    """
    Uses unidecode to normalize text. Requires unidecode package

    Parameters
    ----------
    text : str
      the proposed text with diacritics and so on.

    Returns
    -------
    text : str
      decoded text if unidecode was avail



    Example
    -------
      ```
      str_txt = "Ha ha ha, m\u0103 bucur c\u0103 ai \u00eentrebat!"
      str_simple = self.normalize_text(str_text)
      ```


    """
    text = text.replace('\t', '  ')
    try:
      from unidecode import unidecode
      text = unidecode(text)
    except:
      pass
    return text  
  
  
  def sanitize_name(self, name: str)->str:
    """
    Returns a sanitized name that can be used as a variable name

    Parameters
    ----------
    name : str
        the proposed name

    Returns
    -------
    str
        the sanitized name
    """
    return re.sub(r'[^\w\.-]', '_', name)
  
  def convert_size(self, size, unit):
    """
    Given a size and a unit, it returns the size in the given unit

    Parameters
    ----------
    size : int
        value to be converted
    unit : str
        one of the following: 'KB', 'MB', 'GB'

    Returns
    -------
    _type_
        _description_
    """
    new_size = size
    if unit == ct.FILE_SIZE_UNIT.KB:
      new_size = size / 1024
    elif unit == ct.FILE_SIZE_UNIT.MB:
      new_size = size / 1024**2
    elif unit == ct.FILE_SIZE_UNIT.GB:
      new_size = size / 1024**3
    return new_size  
  
  def managed_lock_resource(self, str_res, condition=True):
    """
    Managed lock resource. Will lock and unlock resource automatically. 
    To be used in a with statement.
    The condition parameter allows users to disable the lock if desired.

    Parameters
    ----------
    str_res : str
      The resource to lock.
    condition : bool, optional
      If False the lock will not be acquired. The default is True.

    Returns
    -------
    LockResource
      The lock resource object.

    Example
    -------
    ```
    with self.managed_lock_resource('my_resource'):
      # do something
    ```

    ```
    # will control if the following operation is locked or not based on this flag
    locking = False
    with self.managed_lock_resource('my_resource', condition=locking):
      # do something
    ```
    """
    return self.log.managed_lock_resource(str_res=str_res, condition=condition)

  def lock_resource(self, str_res):
    """
    Locks a resource given a string. Alias to `self.log.lock_resource`

    Parameters
    ----------
    str_res : str
        the resource name
    """
    return self.log.lock_resource(str_res)

  def unlock_resource(self, str_res):
    """
    Unlocks a resource given a string. Alias to `self.log.unlock_resource`

    Parameters
    ----------
    str_res : str
        the resource name
    """
    return self.log.unlock_resource(str_res)

  def create_numpy_shared_memory_object(self, mem_name, mem_size, np_shape, np_type, create=False, is_buffer=False, **kwargs):
    """
    Create a shared memory for numpy arrays. 
    This method returns a `NumpySharedMemory` object that can be used to read/write numpy arrays from/to shared memory.
    Use this method instead of creating the object directly, as it requires the logger to be set.

    For a complete set of parameters, check the `NumpySharedMemory` class from `core.utils.system_shared_memory`

    Parameters
    ----------
    mem_name : str
        the name of the shared memory
    mem_size : int
        the size of the shared memory. can be ignored if np_shape is provided
    np_shape : tuple
        the shape of the numpy array. can be ignored if mem_size is provided
    np_type : numpy.dtype
        the type of the numpy array
    create : bool, optional
        create the shared memory if it does not exist, by default False
    is_buffer : bool, optional
        if True, the shared memory will be used as a buffer, by default False


    Returns
    -------
    NumPySharedMemory
        the shared memory object
    """
    
    return NumpySharedMemory(
      mem_name=mem_name,
      mem_size=mem_size,
      np_shape=np_shape,
      np_type=np_type,
      create=create,
      is_buffer=is_buffer,
      log=self.log,
      **kwargs
    )
    
    
  def get_temperature_sensors(self, as_dict=True):
    """
    Returns the temperature of the machine if available

    Returns
    -------
    dict
      The dictionary contains the following:
      - 'message': string indicating the status of the temperature sensors
      - 'temperatures': dict containing the temperature sensors
    """
    return self.log.get_temperatures(as_dict=as_dict)

  @property
  def bs4(self):
    """
    Provides access to the bs4 library

    Returns
    -------
      package


    Example
    -------
      ```

      response = self.requests.get(url)
      soup = self.bs4.BeautifulSoup(response.text, "html.parser")
      ```

    """
    return bs4
  
  def get_gpu_info(self, device_id=0):
    """
    Returns the GPU information
    
    Parameters
    ----------
    device_id : int, optional
      The device id. The default is 0.
      
    Returns
    -------
    dict
      The dictionary containing the GPU information
    """
    return self.log.get_gpu_info(device_id=device_id)


  def bytes_to_base64(self, input_bytes, compress=False, url_safe=False):
    """Transforms a bytes object into a base64 encoded string

    Parameters
    ----------
    input_bytes : bytes
        the input bytes
        
    compress : bool, optional
        if True, the bytes will be compressed before encoding. The default is False.

    Returns
    -------
    str: base64 encoded string
    """
    if compress:
      b_code = zlib.compress(input_bytes, level=9)
    else:
      b_code = input_bytes
    if url_safe:
      b_encoded = base64.urlsafe_b64encode(b_code)
    else:
      b_encoded = base64.b64encode(b_code)
    str_encoded = b_encoded.decode('utf-8')
    return str_encoded
  

  def string_to_base64(self, txt, compress=False, url_safe=False):
    """Transforms a string into a base64 encoded string

    Parameters
    ----------
    txt : str
        the input string
        
    compress : bool, optional
        if True, the string will be compressed before encoding. The default is False.

    Returns
    -------
    str: base64 encoded string
    """
    b_text = bytes(txt, 'utf-8')
    return self.bytes_to_base64(
      input_bytes=b_text, url_safe=url_safe, compress=compress
    )

  
  def str_to_base64(self, txt, compress=False, url_safe=False):
    """
    Alias for `string_to_base64`
    """
    return self.string_to_base64(txt, compress=compress, url_safe=url_safe)
  
  
  def dict_in_dict(self, dct1 : dict, dct2 : dict):
    """
    Check if dct1 is in dct2

    Parameters
    ----------
    dct1 : dict
        the first dictionary
    dct2 : dict
        the dictionary where we check if dct1 is contained in 

    Returns
    -------
    bool
        True if dct1 is in dct2
    """
    return self.log.match_template(dct2, dct1)

  # payload handling
  if True:
    def receive_and_decrypt_payload(self, data, verbose=0):
      """
      Method for receiving and decrypting a payload addressed to us.
      
      Parameters
      ----------
      data : dict
          The payload to be decrypted.
          
      verbose : int, optional
          The verbosity level. The default is 0.
          
          
      Returns
      -------
      dict
          The decrypted payload addressed to us.
      """
      # Extract the sender, the data and if the data is encrypted.
      sender = data.get(self.const.PAYLOAD_DATA.EE_SENDER, None)
      is_encrypted = data.get(self.const.PAYLOAD_DATA.EE_IS_ENCRYPTED, False)
      encrypted_data = data.get(self.const.PAYLOAD_DATA.EE_ENCRYPTED_DATA, None)
      # Remove the encrypted data from the payload data if it exists.
      result = {k: v for k, v in data.items() if k != self.const.PAYLOAD_DATA.EE_ENCRYPTED_DATA}

      if is_encrypted and encrypted_data:
        # Extract the destination and check if the data is addressed to us.
        dest = data.get(self.const.PAYLOAD_DATA.EE_DESTINATION, [])
        if not isinstance(dest, list):
          dest = [dest]
        # now we check if the data is addressed to us
        if self.e2_addr not in dest:
          # TODO: maybe still return the encrypted data for logging purposes
          if verbose > 0:
            self.P(f"Payload not for local, dest: {dest}. Ignoring.")
          # endif verbose
          return {}
        # endif destination check

        try:
          # This should fail in case the data was not sent to us.
          str_decrypted_data = self.bc.decrypt_str(
            str_b64data=encrypted_data, str_sender=sender,
            # embed_compressed=True, # we expect the data to be compressed
          )
          decrypted_data = self.json_loads(str_decrypted_data)
        except Exception as exc:
          self.P(f"Error decrypting data from {sender}, to: {dest}:\n{exc}", color='r')
          if verbose > 0:
            self.P(f"Received data:\n{self.dict_to_str(result)}")
          # endif verbose
          decrypted_data = None
        # endtry decryption
        
        if decrypted_data is not None:
          # If the decrypted data is not a dictionary, we embed it in a dictionary.
          # TODO: maybe review this part
          if not isinstance(decrypted_data, dict):
            decrypted_data = {'EE_DECRYPTED_DATA': decrypted_data}
          # endif not dict
          if verbose > 0:
            decrypted_keys = list(decrypted_data.keys())
            self.P(f"Decrypted data keys: {decrypted_keys}")
          # endif verbose
          # Merge the decrypted data with the original data.
          result = {
            **result,
            **decrypted_data
          }
        else:
          if verbose > 0:
            self.P(f"Decryption failed. Returning original data.")
          # endif verbose
        # endif decrypted_data is not None
      # endif is_encrypted
      return result


    def check_payload_data(self, data, verbose=0):
      """
      Method for checking if a payload is addressed to us and decrypting it if necessary.
      Parameters
      ----------
      data : dict
          The payload data to be checked and maybe decrypted.

      verbose : int, optional
          The verbosity level. The default is 0.
      Returns
      -------
      dict
          The original payload data if not encrypted.
          The decrypted payload data if encrypted and the payload was addressed to us.
          None if the payload was encrypted but not addressed to us.
      """
      return self.receive_and_decrypt_payload(data=data, verbose=verbose)
    
    
    def get_hash(self, str_data: str, length=None, algorithm='md5'):
      """
      This method returns the hash of a given string.
      
      Parameters
      ----------
      str_data : str
          The string to be hashed.
      
      length : int, optional
          The length of the hash. The default is None.
          
      algorithm : str, optional
          The algorithm to be used. The default is 'md5'.
          
      Returns
      -------
      
      str
          The hash of the string.
          
      Example
      -------
      
      ```
      hash = plugin.get_hash('test', length=8, algorithm='md5')
      ```
      
      
      """
      assert algorithm in ['md5', 'sha256'], f"Invalid algorithm: {algorithm}"
      bdata = bytes(str_data, 'utf-8')
      if algorithm == 'md5':
        h = hashlib.md5(bdata)
      elif algorithm == 'sha256':
        h = hashlib.sha256(bdata)
      result = h.hexdigest()[:length] if length is not None else h.hexdigest()
      return result


  # Computer vision utils - maybe move to cv package
  if True:      
    def image_entropy(self, image):
      """
      Computes the entropy of an image.

      Parameters
      ----------
      image : cv2 image | PIL image | np.ndarray
          the input image.

      Returns
      -------
      entropy: float
          the entropy of the image
      """

      if image is None:
        # self.P("Image is None")
        return 0

      np_image = np.array(image)
      entropy = 0
      # Build a 1D histogram over pixel intensities (0..255), then normalize
      np_marg = np.histogramdd(np.ravel(np_image), bins=256)[0] / np_image.size
      # Filter out zero-valued bins (log2(0) is undefined)
      np_marg = np_marg[np.where(np_marg > 0)]
      # Apply Shannon entropy: -sum(p * log2(p))
      entropy = -np.sum(np.multiply(np_marg, np.log2(np_marg)))

      return entropy    
    
    @staticmethod
    def rgb_to_lab(rgb):
      """
      Convert an sRGB color (0..255 per channel) to CIE Lab (D65).

      Parameters
      ----------
      rgb : tuple of (R, G, B)
          R, G, B in [0..255].

      Returns
      -------
      lab : tuple of (L, a, b)
          L in [0..100], a in [-128..127], b in [-128..127] (approximately).
      """
      # Reference white point (D65), used in the Lab conversion
      Xn, Yn, Zn = 95.047, 100.000, 108.883

      # 1) Convert sRGB (0..255) to [0..1]
      r, g, b = [channel / 255.0 for channel in rgb]

      # 2) Convert sRGB to linear RGB
      #    If c <= 0.04045 then c/12.92 else ((c+0.055)/1.055)^2.4
      def inv_srgb_comp(c):
          return c / 12.92 if c <= 0.04045 else ((c + 0.055) / 1.055) ** 2.4

      r_lin = inv_srgb_comp(r)
      g_lin = inv_srgb_comp(g)
      b_lin = inv_srgb_comp(b)

      # 3) Linear RGB to XYZ
      #    Matrix from sRGB spec (D65)
      #    X = 0.4124*r_lin + 0.3576*g_lin + 0.1805*b_lin
      #    Y = 0.2126*r_lin + 0.7152*g_lin + 0.0722*b_lin
      #    Z = 0.0193*r_lin + 0.1192*g_lin + 0.9505*b_lin
      X = r_lin * 0.4124 + g_lin * 0.3576 + b_lin * 0.1805
      Y = r_lin * 0.2126 + g_lin * 0.7152 + b_lin * 0.0722
      Z = r_lin * 0.0193 + g_lin * 0.1192 + b_lin * 0.9505

      # Scale to [0..100]
      X *= 100.0
      Y *= 100.0
      Z *= 100.0

      # 4) XYZ to Lab
      #    f(t) = t^(1/3) if t > (6/29)^3 else (t / (3 * (6/29)^2)) + (4/29)
      def f_xyz(t):
          return t ** (1.0 / 3.0) if t > (6.0 / 29.0) ** 3 else (
              (t / (3.0 * (6.0 / 29.0) ** 2)) + (4.0 / 29.0)
          )

      fx = f_xyz(X / Xn)
      fy = f_xyz(Y / Yn)
      fz = f_xyz(Z / Zn)

      L = 116.0 * fy - 16.0
      a = 500.0 * (fx - fy)
      b = 200.0 * (fy - fz)

      return (L, a, b)

    @staticmethod
    def infer_color(rgb, defaults=COMMON_COLORS, scale_factor=10, return_distances=False):
      """
      Classify an input RGB color against a predefined set of colors by picking
      the closest color.

      Parameters
      ----------
      rgb : tuple
          The RGB color to classify, as a tuple of three integers (R, G, B).
          
      defaults : dict, optional
          A dictionary of default colors to classify against. The default is
          a dictionary of common colors.
          The keys are color names and the values are their RGB values.
          The RGB values are expected to be in the range [0, 255].
      
      Returns
      -------
      str
          The name of the color that is closest to the input RGB color.


      """
      color_names = list(defaults.keys())
      color_values = np.array([
          _UtilsBaseMixin.rgb_to_lab(defaults[x]) for x in color_names
        ],
        dtype=np.float32
      )

      # Convert input to a float NumPy array
      color_arr = np.array(_UtilsBaseMixin.rgb_to_lab(rgb), dtype=np.float32)

      # scale_factor = 10
      # color_values = (color_values / scale_factor).round(0)
      # color_arr = (color_arr / scale_factor).round(0)
      # color_values = color_values / (np.sum(color_values, axis=1, keepdims=True) + 1)
      # color_arr = color_arr / (np.sum(color_arr) + 1)

      # Compute the distance
      diffs = color_values - color_arr  # shape (N, 3)
      diffs = diffs**2
      # np.abs(diffs)
      distances = np.sum(diffs, axis=1)

      # Find index of the color with the minimal distance
      idx = np.argmin(distances)
      if return_distances:
        return color_names[idx], distances
      return color_names[idx]


    def classify_color(self, rgb, defaults=COMMON_COLORS):
      """
      Classify an input RGB color against a predefined set of colors by picking
      the closest color.

      Parameters
      ----------
      rgb : tuple
          The RGB color to classify, as a tuple of three integers (R, G, B).
      
      Returns
      -------
      str
          The name of the color that is closest to the input RGB color.
      """
      return self.infer_color(rgb=rgb, defaults=defaults)
  
    def get_crop_color(self, np_rgb_crop, defaults=COMMON_COLORS):
      """
      Get the color of a crop.

      Parameters
      ----------
      np_rgb_crop : np.ndarray
          The input crop, as a NumPy array in [H,W,C] RGB format.

      Returns
      -------
      str
          The name of the color that is closest to the input crop.
      """
      h, w, _ = np_rgb_crop.shape
      crop_ratio = 0.5
      new_h = int(h * crop_ratio)
      new_w = int(w * crop_ratio)
      start_row = (h - new_h) // 2
      start_col = (w - new_w) // 2
      center_crop_rgb = np_rgb_crop[
        start_row:start_row + new_h, 
        start_col:start_col + new_w
      ]
      
      # Compute median across each channel
      median_r = int(np.median(center_crop_rgb[:, :, 0]))
      median_g = int(np.median(center_crop_rgb[:, :, 1]))
      median_b = int(np.median(center_crop_rgb[:, :, 2]))
      median_color_rgb = (median_r, median_g, median_b)

      # Classify using our updated infer_color
      result = self.infer_color(median_color_rgb, defaults=defaults)
      return result

# endclass _UtilsBaseMixin



if __name__ == '__main__':
  from naeural_core import Logger
  from copy import deepcopy

  log = Logger("UTL", base_folder='.', app_folder='_local_cache')

  e = _UtilsBaseMixin()
  e.log = log  
  e.P = print
  
  TEST_D_IN_D = False
  TEST_DICTS = False
  TEST_GIT = False
  
  if TEST_D_IN_D:
    d2 = {
      "SIGNATURE" : "TEST1",
      "DATA" : {
        "A" : 1,
        "B" : 2,
        "C" : {
          "C1" : 10,
          "C2" : 20
        }
      }
    }
    
    d10 = {
      "SIGNATURE" : "TEST1",
      "DATA" : {
        "C" : {
          "C2" : 20
        }
      }
    }
    
    d11 = {
      "SIGNATURE" : "TEST1",
      "DATA" : {
        "C" : {
          "C2" : 1,
        }
      }
    }
    
    log.P("Test 1: d10 in d2: {}".format(e.dict_in_dict(d10, d2)))
    log.P("Test 2: d11 in d2: {}".format(e.dict_in_dict(d11, d2)))
  
  if TEST_DICTS:

    d1 = e.DefaultDotDict(str)
    d1.a = "test"
    print(d1.a)
    print(d1.c)

    d1 = e.DefaultDotDict(lambda: str, {'a' : 'test', 'b':'testb'})
    print(d1.a)
    print(d1.b)
    print(d1.c)
    
    d1c = deepcopy(d1)
    
    d20 = {'k0':1, 'k1': {'k11': 10, 'k12': [{'k111': 100, 'k112':200}]}}
    d2 = e.NestedDotDict(d20)
    d20c = deepcopy(d20)
    d2c = deepcopy(d2)
    
    print(d2)
    print(d2.k0)
    print(d2.k1.k12[0].k112)
    
    
    
    d3 = defaultdict(lambda: DefaultDotDict({'timestamp' : None, 'data' : None}))
    
    s = json.dumps(d20)
    print(s)
    
    b64 = e.string_to_base64(s)
    print("{}: {}".format(len(b64), b64[:50]))
    print(e.base64_to_str(b64))

    b64c = e.string_to_base64(s, compress=True)
    print("{}: {}".format(len(b64c), b64c[:50]))
    print(e.base64_to_str(b64c, decompress=True))
      
    config = e.load_config_file(fn='./config_startup.txt')
    

    d4 = NestedDefaultDotDict()
    
    assert d4.test == {}, "Accessing an undefined key did not return empty dict."
    
    # Test case 2: Automatically creates nested dictionaries and sets value
    d4.test2.x = 5
    assert d4.test2.x == 5, "Nested assignment failed."
    
    # Test case 3: Auto-creates both test3 and test4, where test4 has value None
    _ = d4.test3.test4  # Access to create
    assert len(d4.test3) != 0 and len(d4.test3.test4) == 0, "Nested auto-creation failed."
    
    print("All tests passed.")
  
  if TEST_GIT:
    repo = 'https://github.com/Ratio1/edge_node_launcher'
    sub_folder = "test_repo"
    full_path = os.path.join(log.get_output_folder(), sub_folder)
    
    
    # cloning
    local_path = e.git_clone(repo_url=repo, repo_dir=sub_folder, pull_if_exists=True)
    # remote hash
    remote_hash = e.git_get_last_commit_hash(repo_url=repo)
    # local hash
    local_hash = e.git_get_local_commit_hash(repo_dir=local_path)
    # output
    log.P(f"Cloned to: <{local_path}>")
    log.P(f"Remote: <{remote_hash}>")
    log.P(f"Local:  <{local_hash}>")
  
  
  colors_dict = {
      "red": (255, 0, 0),
      "dark_red": (100, 0, 0),
      "purple": (120, 0, 120),
      "blue_purple": (125, 74, 250),
      "blue_cyan": (17, 82, 147),
      "cyan_blue_green": (0, 252, 206),
      "black_brown": (45, 44, 45),
      "brown_red_1": (135, 65, 30),
      "brown_red_2": (90, 20, 4),
      "yellow_gold_1": (233, 211, 6),
      "yellow_gold_2": (203, 185, 18),
      "green_blue_1": (57, 111, 80),
      "green_blue_2": (59, 110, 86),
      "green_blue_3": (37, 99, 75),
      "green_blue_4": (58, 110, 86),
  }

  
  for n, c in colors_dict.items():
    name, distances = e.infer_color(c, scale_factor=20, return_distances=True)
    dist = [round(x,2) for x in distances]
    log.P(f"{n}:{c} => {name}: {dist}")


import json

from naeural_core.bc import DefaultBlockEngine
from naeural_core import Logger


if __name__ == '__main__':
  l = Logger('AENC', base_folder='.', app_folder='_local_cache')
  
  snd_name = "asym-enc-dec-1"
  sender = DefaultBlockEngine(
    log=l,
    name=snd_name,
    config= {
      "PEM_FILE"     : snd_name + ".pem",
      "PASSWORD"     : None,      
      "PEM_LOCATION" : "data"
    },
  )
  
  data = {
    "value" : "Hello World",
  }
  str_data = json.dumps(data)
  receiver_address = "0xai_A3vtcVIv_yL7k945IuhNjLUXKj2DPvbapoH4D6ZairfT"
  
  l.P("Sender <{}> private data: {}".format(sender.address, data))
  enc_data = sender.encrypt(
    plaintext=str_data,
    receiver_address=receiver_address,
  )
  l.P("Sender <{}> encrypted data: {}".format(sender.address, enc_data))
  
  payload = {
    "EE_ENCRYPTED_DATA" : True,
    "ENCRYPTED_DATA" : enc_data,    
  }  
  
  signature = sender.sign(
    dct_data=payload
  )
  
  out_payload = payload
  
  str_output_payload = json.dumps(out_payload)
  print("Sender <{}> outgoing payload signed with {}:\n {}".format(
    sender.address, signature, json.dumps(out_payload, indent=2)))    
    


import torch as th
import json
import os

from naeural_core.utils.tracing.base_scripter import BaseScripter
from ratio1 import load_dotenv


load_dotenv()


def load_model_full(
  log, model_name, model_factory, weights, hyperparams,
  weights_filename=None, hyperparams_filename=None,
  device='cpu', return_config=False
):
  """
  Method for loading a model from a given class, with given configuration
  (either as a dictionary or as a .json) hyperparameters.
  Parameters
  ----------
  log - Logger, swiss knife object used in all DecentrAI
  model_name - str, name of the model
  model_factory - class extending torch.nn.Module, class of the given model
  weights - str, local path or url to file with the weights for the model
  device - str or torch.device, device on which the input will be put
  hyperparams - str or dict
  - if dict it will be considered the configuration of the specified model
  - if str it will be considered either a local path or an url to a file containing the hyperparameters
  weights_filename - str or None
  - if weights is an url this has to be the filename to be used when saving the file locally
  hyperparams_filename - str or None
  - if hyperparams is an url this has to be the filename to be used when saving the file locally
  return_config - bool, whether to return the config as a dictionary

  Returns
  -------
  res - (torch.nn.Module, dict) if return_config else torch.nn.Module
  The instantiated model with the weights loaded and if required, the config dictionary of the model.
  """
  log.P(
    f'Attempting to load model {model_name} with config: {hyperparams} and weights at: {weights} on device {device}')
  if weights_filename is None and os.path.isfile(weights):
    weights_filename = os.path.split(weights)[-1]
  elif weights_filename is None:
    raise Exception('`weights_filename` not provided and `weights` is not a local file. Please either provide '
                    'a path to an existing local file or specify `weights_filename`!')
  if isinstance(hyperparams, str) and hyperparams_filename is None and os.path.isfile(hyperparams):
    hyperparams_filename = os.path.split(hyperparams)[-1]
  elif isinstance(hyperparams, str) and hyperparams_filename is None:
    raise Exception('`hyperparams_filename` not provided and `hyperparams` is not a local file. Please either provide '
                    'a path to an existing local file or specify `hyperparams_filename`!')

  download_kwargs = log.config_data.get('MODEL_ZOO_CONFIG', {
      "endpoint": os.environ["EE_MINIO_ENDPOINT"],
      "access_key": os.environ["EE_MINIO_ACCESS_KEY"],
      "secret_key": os.environ["EE_MINIO_SECRET_KEY"],
      "secure": os.environ["EE_MINIO_SECURE"],
      "bucket_name": "model-zoo"
    })

  saved_files, msg = log.maybe_download_model(
    url=weights,
    model_file=weights_filename,
    **download_kwargs,
  )

  hyperparams_files = [None]
  if type(hyperparams) == str:
    hyperparams_files, _ = log.maybe_download_model(
      url=hyperparams,
      model_file=hyperparams_filename,
      **download_kwargs
    )
    hyperparams = None
  model, hyperparams = log.load_model_and_weights(
    model_class=model_factory,
    weights_path=saved_files[0],
    config_path=hyperparams_files[0],
    return_config=True,
    device=device,
    config=hyperparams
  )
  model.eval()

  return (model, hyperparams) if return_config else model

class BaseTorchScripter(BaseScripter):
  def __init__(
    self, log, model, model_name, input_shape,
    model_config=None, preprocess_method=None,
    matching_method=None,
    predict_method=None, use_fp16=False,
    use_amp=False,
    gold_model=None
  ):
    self.extension = '.ths'
    super(BaseTorchScripter, self).__init__(
      log, model, model_name, input_shape,
      model_config, preprocess_method,
      matching_method,
      predict_method, use_fp16,
      use_amp, gold_model)
    return

  def load(self, fn_path, device, batch_size, use_fp16):
    extra_files = {'config.txt': ''}
    ts_model = th.jit.load(
      f=fn_path,
      map_location=device,
      _extra_files=extra_files,
    )
    config = json.loads(extra_files['config.txt'].decode('utf-8'))
    if use_fp16:
      ts_model.half()
    return ts_model, config

  def default_predict(self, model, inputs):
    return model.predict(inputs) if hasattr(model, 'predict') else model(inputs)

  def convert(self, inputs, config, fn):
    if hasattr(self.model, 'predict'):
      try:
        traced_model = th.jit.script(self.model)
      except Exception as e:
        if not isinstance(inputs, dict):
          inputs = {"forward": inputs, "predict": inputs}
        # endif inputs not already a dict
        traced_model = th.jit.trace_module(
          self.model,
          inputs,
          strict=False
        )
    else:
      traced_model = th.jit.trace(self.model, inputs, strict=False)

    extra_files = {'config.txt': json.dumps(config)}
    self.log.P(f"  Saving '{fn}'...")
    traced_model.save(fn, _extra_files=extra_files)
    return traced_model




class _AlertWitnessMixin(object):
  def __init__(self):
    """
    self.last_witness - dictionary with the last alert-able and non alert-able
    witness image in each alerter
    self.last_witness[alerter][0] - data about last non alert-able witness image based on the specified alerter
    self.last_witness[alerter][1] - data about last alert-able witness image based on the specified alerter
    """
    self.__last_witness_kwargs = {}
    super(_AlertWitnessMixin, self).__init__()
    return

  """
  Update data for the current or any other state of the specified alerter
  """

  def alerter_get_current_frame_state(self, observation, alerter='default'):
    """
    This function returns the possible next alerter position based on the current alerter state and the current observation.

    If the current observation can change the alerter state from A to B, the function returns the position of the state B.
    (this ensures that an alertable observation will be saved to the alertable state, no matter the current alerter state)

    If the current observation cannot change the alerter state from A to B, the function returns the position of the state A.

    Parameters
    ----------
    observation : float
        The current observation

    alerter : str, optional
        The alerter name, by default 'default'

    Returns
    -------
    int
        The possible next alerter position
    """
    pos = None
    # we use this instead of self.cfg_alert_raise_value and self.cfg_alert_lower_value
    # because we may have multiple alerters with different configurations
    dct_alerter_config = self.get_alerter(alerter=alerter).get_setup_values()

    if observation is not None:
      # if the current observation can change the alerter state from A to B
      # we save the observation to state B
      # (we do not care about the current alerter state here)
      if observation > dct_alerter_config['raise_alert_value']:
        pos = 1
      elif observation < dct_alerter_config['lower_alert_value']:
        pos = 0

    if pos is None:
      # if the current observation cannot change the alerter status from A to B
      # we save the observation to state A
      # (the current alerter state)
      if self.alerter_is_alert():
        pos = 1
      else:
        pos = 0
    return pos

  def update_witness_kwargs(self, witness_args, pos, alerter='default'):
    if alerter not in self.__last_witness_kwargs.keys():
      # instantiating the values for the current alerter if no data
      # was previously provided
      self.__last_witness_kwargs[alerter] = [None, None]
    self.__last_witness_kwargs[alerter][pos] = witness_args

    return

  """
  Get data for the current or any other state of the specified alerter
  """

  def get_current_witness_kwargs(self, pos=None, alerter='default', demo_mode=False):
    if pos is None:
      if demo_mode:
        pos = self.alerter_get_current_frame_state(
          observation=self.alerter_get_last_value(alerter=alerter),
          alerter=alerter
        )
      else:
        pos = int(self.alerter_is_alert(alerter=alerter))
    witness_args = self.__last_witness_kwargs[alerter][pos]
    # if witness_args is None:
    #   return {}
    return witness_args

  """
  Get generated witness from data provided by get_current_witness_kwargs
  """

  def get_current_witness(self, pos=None, alerter='default'):
    witness_args = self.get_current_witness_kwargs(pos=pos, alerter=alerter)
    return self.get_witness_image(**witness_args) if witness_args is not None else None



import requests

_SERVER = 'https://home.sensibo.com/api/v2'

class SensiboClientAPI(object):
  def __init__(self, api_key):
    self._api_key = api_key

  def _get(self, path, ** params):
    params['apiKey'] = self._api_key
    response = requests.get(_SERVER + path, params=params)
    response.raise_for_status()
    return response.json()

  def _patch(self, path, data, **params):
    params['apiKey'] = self._api_key
    response = requests.patch(_SERVER + path, params=params, data=data)
    response.raise_for_status()
    return response.json()

  def devices(self):
    result = self._get('/users/me/pods', fields='id,room')
    return {x['room']['name']: x['id'] for x in result['result']}

  def pod_measurement(self, pod_uid):
    result = self._get('/pods/{}/measurements'.format(pod_uid))
    return result['result']


if __name__ == "__main__":
  api_key = '0B073b470DeXHoqmXmdeBpVzBbHcLh'
  device_name = "Alex's device"
  
  client = SensiboClientAPI(api_key)
  devices = client.devices()
  print("-" * 10, "devices", "-" * 10)
  print(devices)

  uid = devices[device_name]
  lst_measurements = client.pod_measurement(uid)
  for measurement in lst_measurements:
    for k,v in measurement.items():
      print('{}: {}'.format(k, v))


import urllib.request
import ssl
import json
import uuid
import shutil
import os
import traceback
import requests

from naeural_core import Logger
from naeural_core.config.base import BaseConfigRetrievingPlugin

import naeural_core.constants as ct

_CONFIG = {
  **BaseConfigRetrievingPlugin.CONFIG,
  'VALIDATION_RULES': {
    **BaseConfigRetrievingPlugin.CONFIG['VALIDATION_RULES'],
  },
}

class RemoteConfigRetriever(BaseConfigRetrievingPlugin):
  CONFIG = _CONFIG
  def __init__(self, log : Logger,
               **kwargs):
    super(RemoteConfigRetriever, self).__init__(log=log, prefix_log='[RemoteCR]', **kwargs)
    return
  
  def _connect(self, **kwargs):
    self.P("  'Fake' connect - directly accessing url...")
    return

  def _urlopen_config(self, url):
    TIMEOUT = 10
    config_fallback = self.config.get(ct.CONFIG_APP_FALLBACK)
    self.P("  Running '{}'...".format(url))
    try:
      if False:
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
      ssl._create_default_https_context = ssl._create_unverified_context  
      with urllib.request.urlopen(url, timeout=TIMEOUT) as f:
        content = f.read().decode('utf-8')
        config = json.loads(content)
    except Exception as exc:
      self.P("Failed urllib request:\n{}\nConfig: {}".format(exc, self.config), color='r')
      try:
        resp = requests.get(url, verify=False, timeout=TIMEOUT)
        content = resp.content.decode('utf-8')
        config = json.loads(content)
      except Exception as exc:
        self.P("`requests` also failed after urllib: {}".format(exc), color='r')
        if config_fallback is not None:
          self.P("Using fallback config {}...".format(config_fallback))
          self.P(" File {}".format(
            "is valid." if os.path.isfile(config_fallback) else "DOES NOT exist!"
          ))
          with open(config_fallback, 'rt') as fh:
            config = json.load(fh)
          self.P("  Loaded fallback config.")
        else:
          raise ValueError("Failed urllib request: {}".format(exc))
    return config

  def _get_app_configuration(self, endpoint):
    assert isinstance(endpoint, str)
    return self._urlopen_config(endpoint)

  def _get_streams_configurations(self, endpoint):
    if isinstance(endpoint, str):
      ### streams can be given as a link to a folder with multiple files
      return self.__handle_folder_streams_configurations(endpoint)
    elif isinstance(endpoint, list):
      ### streams can be given as list of links to separate streams configuration files
      return self.__handle_multiple_files_streams_configurations(endpoint)
    elif isinstance(endpoint, dict):
      ### streams can be given as a dictionary where the key is "FOLDERS" and each element is a folder with multiple files
      return self.__handle_multiple_folders_streams_configurations(endpoint)
    else:
      raise ValueError("Unknown value {} for parameter `endpoint`".format(endpoint))

  def __handle_multiple_folders_streams_configurations(self, endpoint):
    folders = endpoint['FOLDERS']
    lst_config_streams = []
    for fld in folders:
      lst_config_streams += self.__handle_folder_streams_configurations(fld)
    return lst_config_streams

  def __handle_folder_streams_configurations(self, endopoint):
    _tmp_fld = str(uuid.uuid4())

    saved, _ = self.log.maybe_download(
      url=endopoint,
      fn=_tmp_fld,
      target='data',
      unzip=True,
      verbose=False,
      print_progress=False
    )

    save_dir = saved[0]

    lst_config_streams = []
    for fn in os.listdir(save_dir):
      if fn.endswith('.txt') or fn.endswith('.json'):
        try:
          lst_config_streams.append(self.log.load_json(
            fname=os.path.join(save_dir, fn),
            folder=None,
            verbose=False
          ))
        except Exception as e:
          self.P("JSON syntactic error with stream {}\n{}".format(fn, e), color='r')

    shutil.rmtree(save_dir)

    return lst_config_streams

  def __handle_multiple_files_streams_configurations(self, endpoint):
    lst_config_streams = []
    for url in endpoint:
      lst_config_streams.append(self._urlopen_config(url))

    return lst_config_streams



from ratio1 import BaseDecentrAIObject

import numpy as np
import matplotlib.pyplot as plt

from sklearn.cluster import KMeans
from collections import Counter

class ElbowKmeans(BaseDecentrAIObject):
  def __init__(self, log, feats, min_nr_clusters=2, max_nr_clusters=30):
    if log is None or (type(log).__name__ != 'Logger'):
      raise ValueError("Loggger object is invalid: {}".format(log))
    super().__init__(log=log)
    
    self.X = feats
    self.errors  = []
    self.min_nr_clusters = min_nr_clusters
    self.max_nr_clusters = max_nr_clusters
    self.P("Initialized SearchKmeans class with the following params:")
    self.P(" * feats shape:     {}".format(feats.shape))
    self.P(" * min_nr_clusters: {}".format(min_nr_clusters))
    self.P(" * max_nr_clusters: {}".format(max_nr_clusters))
    return


  def _dist_line(self, x_A, y_A, x_B, y_B, x0, y0):
    m = (y_A - y_B) / (x_A - x_B)
    b = (y_B * x_A - y_A * x_B) / (x_A - x_B)
    
    distance = np.abs(m * x0 - y0 + b) / np.sqrt(1 + m**2);
    return distance
    
    
  def _check_stop_condition(self):
    if len(self.errors) < 3:
      return False, None

    sol_iter = []
    for i in range(len(self.errors)-2):
      tmp_error = self.errors[0:i+3]

      sol_x = 0
      d_max = 0

      for j in range(1, len(tmp_error)-1):
        dist = self._dist_line(0, tmp_error[0], i+2, tmp_error[i+2], j, tmp_error[j])

        if dist > d_max:
          d_max = dist
          sol_x = j
        #endif
      #endfor
      sol_iter.append(sol_x)
    #endfor 
    
    histogram = Counter(sol_iter)
    no_of_occurence = 5
    if no_of_occurence in histogram.values():
      return True, list(histogram.keys())[list(histogram.values()).index(no_of_occurence)]
    else:
      return False, None
   
      
  def get_optimum_nr_clusters(self, n_init=5, max_iter=300, verbose=0):
    self.P("Finding the optimum nr of clusters ...")
    self.P("KMeans params:")
    self.P(" * n_init={}".format(n_init))
    self.P(" * max_iter={}".format(max_iter))
    self.P(" * verbose={}".format(verbose))
    range_ = range(self.min_nr_clusters, self.max_nr_clusters)
    for i, nr_clusters in enumerate(range_):
      self.P("Iteration {}/{} - {} clusters".format(i+1, len(range_), nr_clusters))
      self.kmeans_object = KMeans(nr_clusters, n_init=n_init,
                                  max_iter=max_iter, verbose=verbose)
      self.kmeans_object.fit(self.X)
      self.P(" ", t=True)
      self.errors.append(self.kmeans_object.inertia_)
      ret, iter_sol = self._check_stop_condition()
      if ret is True:
        optimum_nr = iter_sol + self.min_nr_clusters
        self.kmeans_object = KMeans(optimum_nr)
        self.kmeans_object.fit(self.X)

        plt.figure()
        plt.plot(np.arange(len(self.errors)) + self.min_nr_clusters, self.errors)
        plt.scatter(optimum_nr, self.errors[iter_sol])
        self.P(" the system found {} optimum clusters.".format(optimum_nr))
        return optimum_nr
    
  def get_labels(self):
    return self.kmeans_object.labels_


"""

TODO: What is the purpose of LOCAL_COMMS_01 ?


"""

from naeural_core.business.base import BasePluginExecutor

__VER__ = '1.0.0'

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'LOCAL_COMMS_ENABLED_ON_STARTUP': False,
  'ALLOW_EMPTY_INPUTS': True,

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}


class LocalComms01Plugin(BasePluginExecutor):
  CONFIG = _CONFIG

  def on_init(self):
    self.__comm_manager = self.global_shmem["comm_manager"]
    # state of the local comms
    self.__enabled = False

    # current request: True to enable, False to disable, None to do nothing
    self.__current_request = self.cfg_local_comms_enabled_on_startup

    # local MQTT broker process
    self.__local_mqtt_broker = None
    return

  def on_command(self, data, enable=False, disable=False, **kwargs):
    if (isinstance(data, str) and data.upper() == 'ENABLE') or enable:
      self.__current_request = True
    elif (isinstance(data, str) and data.upper() == 'DISABLE') or disable:
      self.__current_request = False
    return


  def _start_local_mqtt_broker(self):
    import subprocess as sp
    
    # create the config file
    if not self.os_path.exists('mosquitto.conf'):
      with open('mosquitto.conf', 'w') as f:
        f.write("listener 1883\n")
        f.write("allow_anonymous true")
    # endif

    # mosquitto -p 1883 -c mosquitto.conf
    CMD = [
      'mosquitto',
      # '-p', '1883',
      '-c', 'mosquitto.conf'
    ]
    self.P(f"Starting local MQTT broker process via command: {CMD}")
    self.__local_mqtt_broker = sp.Popen(CMD)

    self.P("Local MQTT broker process starting...")

    self.sleep(3)
    # i want to check if the process started and is running
    if self.__local_mqtt_broker.poll() is None:
      self.P("Local MQTT broker process is running.")
      return True

    self.P("Local MQTT broker process is not running.", color='r')
    return False

  def _stop_local_mqtt_broker(self):
    self.P("Local MQTT broker process stopping...")
    self.__local_mqtt_broker.terminate()

    self.sleep(3)

    if self.__local_mqtt_broker.poll() is not None:
      self.__local_mqtt_broker = None
      self.P("Local MQTT broker process stopped.")
      return True

    self.P("Local MQTT broker process is still running. Force stopping it now")
    self.__local_mqtt_broker.kill()

    self.sleep(3)
    if self.__local_mqtt_broker.poll() is not None:
      self.__local_mqtt_broker = None
      self.P("Local MQTT broker process stopped.")
      return True

    self.P("Local MQTT broker process is still running.", color='r')
    return False


  def process(self):
    # if no request, do nothing
    if self.__current_request in [None, False]:
      return

    # if not enabled and request is to enable, enable
    if not self.__enabled and self.__current_request:
      started = self._start_local_mqtt_broker()
      if started:
        self.__comm_manager.enable_local_communication()
        self.__enabled = True

    # if enabled and request is to disable, disable
    if self.__enabled and not self.__current_request:
      self.__comm_manager.disable_local_communication()
      self.__enabled = False
      self._stop_local_mqtt_broker()

    # reset request
    self.__current_request = None
    return



import sys
from pathlib import Path

import wandb

FILE = Path(__file__).resolve()
ROOT = FILE.parents[3]  # YOLOv5 root directory
if str(ROOT) not in sys.path:
    sys.path.append(str(ROOT))  # add ROOT to PATH

from train import train, parse_opt
from utils.general import increment_path
from utils.torch_utils import select_device
from utils.callbacks import Callbacks


def sweep():
    wandb.init()
    # Get hyp dict from sweep agent
    hyp_dict = vars(wandb.config).get("_items")

    # Workaround: get necessary opt args
    opt = parse_opt(known=True)
    opt.batch_size = hyp_dict.get("batch_size")
    opt.save_dir = str(increment_path(Path(opt.project) / opt.name, exist_ok=opt.exist_ok or opt.evolve))
    opt.epochs = hyp_dict.get("epochs")
    opt.nosave = True
    opt.data = hyp_dict.get("data")
    device = select_device(opt.device, batch_size=opt.batch_size)

    # train
    train(hyp_dict, opt, device, callbacks=Callbacks())


if __name__ == "__main__":
    sweep()



from collections import deque
from math import sin

from naeural_core.utils.sanchez_utils.Actor import *

BABY_SNAKE_STEPS_LIMIT = 5
SELF_EATING = False


class Snake(Actor):
  def __init__(self, name='', color=None, actor_id=1,
               x=0, y=0, alpha=0, beta=0,
               speed=0, period=0, steps=0,
               life_state=0, length=1, queue=None,
               extra_lifes=0, baby_steps=0
               ):
    super(Snake, self).__init__(
      name=name, color=color, actor_id=actor_id,
      x=x, y=y, alpha=alpha, beta=beta,
      speed=speed, period=period, steps=steps,
      life_state=life_state, baby_steps=baby_steps
    )

    self.length = length
    self._queue = deque([(self.x, self.y)]) if queue is None else queue
    self.extra_lifes = extra_lifes
    self.to_move = 0

  def get_species(self):
    return TYPE_SNAKE

  def get_data_string(self):
    return super(Snake, self).get_data_string() + f', {self.length}'

  def _get_baby_steps_limit(self):
    return BABY_SNAKE_STEPS_LIMIT

  def retire(self):
    super(Snake, self).retire()
    self._queue = None
    self.length = 1
    self.to_move = 0

  def clear_map(self, mat):
    for (x, y) in self._queue:
      mat[x, y] = 0

  def make_baby(self, env, x, y):
    baby = Snake(name=self.name + 'son', color=self.color, x=x, y=y)
    env.add_actor(baby)
    print(f'Baby {baby.get_data_string()} was born!')

  def lose_life(self, mat):
    if self.extra_lifes > 0:
      self.extra_lifes -= 1
    else:
      self.clear_map(mat)
      self.retire()

  def manage_collision(self, actor, env, x, y):
    if type(actor) == type(self):
      if actor.life_state == BABY_STATE or actor.life_state == RETIRED_STATE:
        return
      if self.length >= actor.length:
        actor.clear_map(env.get_current_map())
        actor.retire()
        self.length += 1
      elif self.length < actor.length / 2:
        self.clear_map(env.get_current_map())
        self.retire()
        actor.length += 1
      else:
        self.make_baby(env, x, y)
        self._change_direction()
    elif actor.get_species() == TYPE_HEDGEHOG:
      if actor.life_state == ADULT_STATE:
        return
      if actor.life_state == BABY_STATE:
        actor.retire()
        self.extra_lifes += 1
        return

      self.lose_life(env.get_current_map())
      self._change_direction()

  def get_points(self, h=None, w=None):
    if self._queue is not None and len(self._queue) > 0:
      return list(self._queue)
    return super(Snake, self).get_points(h, w)

  def _future_position(self, speed=None):
    if not hasattr(self, 'rounds_lived'):
      self.rounds_lived = 0
    else:
      self.rounds_lived += 1
      
    if speed is None:
      speed = self.speed
    x, y = super()._future_position(speed)
    
    # now jiggle
    x += self.beta * sin(0.4 * (speed * self.rounds_lived))
    y -= self.alpha * sin(0.4 * (speed * self.rounds_lived))
    return x, y

  def forward(self, h=None, w=None, env=None):
    if self._check_baby_state():
      return

    if self.life_state == RETIRED_STATE:
      return

    self.to_move = self.to_move + self.speed
    steps_to_make = int(self.to_move)
    self.to_move = self.to_move - steps_to_make

    points_to_free = len(self._queue) + steps_to_make - self.length
    if points_to_free > len(self._queue):
      points_to_free = points_to_free - len(self._queue)
      removed_points = [self._queue.popleft() for _ in range(len(self._queue))]
    else:
      points_to_free = max(points_to_free, 0)
      removed_points = [self._queue.popleft() for _ in range(int(points_to_free))]
      points_to_free = 0

    if env is not None:
      mat = env.get_current_map()
      for (i, j) in removed_points:
        mat[i, j] = 0

    added_points = []
    for _ in range(steps_to_make):
      if points_to_free > 0:
        if len(self._queue) > 0:
          (i, j) = self._queue.popleft()
          removed_points.append((i, j))
          points_to_free -= 1
          if env is not None:
            env.get_current_map()[i, j] = 0

      x, y = self._future_position(speed=1)
      x, y = self.still_in_map(x, y, h, w, 1)

      if x is None:
        return

      if env is not None:
        mat = env.get_current_map()
        p_x, p_y = self.matrix_position(h, w, x, y)

        current_value = mat[p_x, p_y]
        if current_value != 0:
          if current_value > 0:
            # actor here
            if current_value == self.id:
              if SELF_EATING:
                # self intersection
                self.clear_map(mat)
                self.retire()
                return
            else:
              actor = env.get_actor(current_value)
              self.manage_collision(actor, env, p_x, p_y)
          else:
            # stone here
            self.clear_map(mat)
            self.retire()

        if self.life_state == RETIRED_STATE:
          return
      # endif
      self._step_forward(1)
      added_points.append(self.matrix_position(h, w))
    # endfor

    for point in added_points[points_to_free:]:
      self._queue.append(point)

    if env is not None:
      mat = env.get_current_map()
      for point in added_points[points_to_free:]:
        mat[point] = self.id













#global dependencies
import os
import inspect  
import importlib
import traceback
import pandas as pd

#local dependencies
from naeural_core import DecentrAIObject

class ModelBenchmarker(DecentrAIObject):
  def __init__(self, log, module_name, class_name, batch_sizes=None, **kwargs):
    super().__init__(log=log, **kwargs)
    assert isinstance(module_name, str), 'Please provide path to module'
    assert isinstance(class_name, str), 'Please provide name of class'
    module_name = module_name.replace('/', '.')
    self._module_name = module_name
    self._class_name = class_name    
    self._kwargs = kwargs
    
    self._batch_sizes = [2**p for p in range(7)]
    if isinstance(batch_sizes, list):
      assert len(batch_sizes) > 0, 'You specified an empty batch sizes list'
      self._batch_sizes = batch_sizes
    return
  
  def startup(self):
    super().startup()    
    self._info = {
      'STEP': [],
      'TOTAL_GB': [],
      'USED_GB': [],
      'TIME': []
      }
    self._cls = None
    return
  
  def _add_info(self, step_name):
    total_gpu_mem = ''
    allocated_gpu_mem = ''
    try:
      gpu_info = self.log.gpu_info()
      total_gpu_mem = gpu_info[0]['TOTAL_MEM']
      allocated_gpu_mem = gpu_info[0]['ALLOCATED_MEM']
    except:
      pass
    timer = self.log.get_timer(step_name)
    total_time = timer.get('END', 0) - timer.get('START', 0)
    
    self._info['STEP'].append('{}_{}'.format(self._class_name, step_name))
    self._info['TOTAL_GB'].append(total_gpu_mem)
    self._info['USED_GB'].append(allocated_gpu_mem)
    self._info['TIME'].append(total_time)
    return
  
  def _start(self):
    step_name = 'START'
    self.log.start_timer(step_name)    
    #no op, just take a snapshot of system status
    self.log.stop_timer(step_name)    
    self._add_info(step_name)
    return
  
  def _import(self):
    step_name = 'IMPORT'
    self.log.start_timer(step_name) 
    _cls = None
    try:
      module = importlib.import_module(self._module_name)
      classes = inspect.getmembers(module, inspect.isclass)
      for obj in classes:
        if obj[0] == self._class_name:
          _cls = obj[1]
          break
      #endfor
      
      if _cls is None:
        raise ValueError('Could not find specified class `{}` in module `{}`'.format(self._class_name, self._module_name))
      
      self._cls = _cls(
        log=self.log,
        **self._kwargs
        )
    except:
      self.P('Exception while loading requested class: {}'.format(traceback.format_exc()))
    finally:
      self.log.stop_timer(step_name)    
      if self._cls is not None:
        self._add_info(step_name)
    return
  
  def _load(self):
    step_name = 'LOAD'
    self.log.start_timer(step_name)
    self._cls.load()
    self.log.stop_timer(step_name)
    self._add_info(step_name)
    return
  
  def _prepare(self):
    step_name = 'PREPARE'
    self.log.start_timer(step_name)
    self._cls.prepare()
    self.log.stop_timer(step_name)
    self._add_info(step_name)
    return
  
  def _draw(self, path_out, inputs, preds):
    self._cls.draw(path_out, inputs, preds)
    return
  
  def _report(self, max_batch_size):
    self.log.p('Creating report ...')
    
    self.log.p('Max batch size: {}'.format(max_batch_size))
    df = pd.DataFrame(self._info)
    self.log.p('\n\n{}'.format(df))
    return df
  
  def run(self, inputs, **kwargs):    
    self._import()
    if self._cls is None:
      self.log.p('Benchmark will not run as requested module/class was not found!')
      return
    
    self._load()
    self._prepare()
    
    max_bs = None
    for bs in self._batch_sizes:
      try:
        step_name = 'PREDICT_BS_{}'.format(bs)
        n_batches = len(inputs) // bs
        if n_batches == 0:
          self.log.p('No batches for bs {}'.format(bs))
          continue
        
        path_out = os.path.join(
          self.log.get_output_folder(),
          self._class_name,
          'BS_{}'.format(bs)
          )
        os.makedirs(path_out, exist_ok=True)
        self.log.p('Infering on bs {}. Number of batches: {}'.format(bs, n_batches))
        for n_batch in range(n_batches):
          _start = n_batch * bs
          _stop = (n_batch + 1) * bs
          batch = inputs[_start:_stop]
          
          self.log.start_timer(step_name)
          preds = self._cls.predict(batch)
          self.log.stop_timer(step_name)          
          
          self._draw(path_out, batch, preds)
        #endfor
        max_bs = bs
        self._add_info(step_name)
      except:
        self.log.p('Exception while running bs {}: {}'.format(bs, traceback.format_exc()))
    #endfor
    
    df = self._report(
      max_batch_size=max_bs
      )
    return df
    
  
    
    
    
    
    
    
    
    
    
    
    
    
    
    
  
    
    
  
  
  


from naeural_core.business.base.drivers.device import Device


class HttpDevice(Device):
  def __init__(self, **kwargs):
    super(HttpDevice, self).__init__(**kwargs)
    self._headers = None
    return

  def device_get(self, path, headers=None, timeout=10, debug=False):  # todo: prefix device
    """
    Get data from device
    Parameters
    ----------
    path
    headers
    timeout

    Returns
    -------

    """
    return self.__execute_request(
      method="GET", 
      path=path, headers=headers, 
      timeout=timeout, debug=debug,
    )

  def device_post(self, path, body, headers=None, timeout=10, debug=False):
    """
    Post data to device
    Parameters
    ----------
    path
    body
    headers
    timeout

    Returns
    -------

    """
    return self.__execute_request(
      method="POST", 
      path=path, body=body, headers=headers, 
      timeout=timeout, debug=debug,
    )

  def device_put(self, path, body, headers=None, timeout=10, debug=False):
    """
    Put data to device
    Parameters
    ----------
    path
    body
    headers
    timeout

    Returns
    -------

    """
    return self.__execute_request(
      method="PUT", 
      path=path, body=body, headers=headers, 
      timeout=timeout, debug=debug,
    )

  def device_delete(self, path, headers=None, timeout=10, debug=False):
    """
    Delete data from device
    Parameters
    ----------
    path
    headers
    timeout

    Returns
    -------

    """
    return self.__execute_request(
      method="DELETE", 
      path=path, headers=headers, 
      timeout=timeout, debug=debug,
    )

  def __execute_request(self, method, path, body=None, headers=None, timeout=10, debug=False):
    """
    Execute request to device
    Parameters
    ----------
    method
    path
    body
    headers
    timeout

    Returns
    -------

    """
    url, dct_headers = self.__get_url_and_headers(headers=headers, path=path)
    if debug:
      self.P(f"Executing {method} request to {url} with headers: {dct_headers} and body: {body}")
    try:
      if method.upper() == "GET":
        response = self.requests.get(url=url, headers=dct_headers, timeout=timeout)
      elif method.upper() == "POST":
        response = self.requests.post(url=url, json=body, headers=dct_headers, timeout=timeout)
      elif method.upper() == "PUT":
        response = self.requests.put(url=url, json=body, headers=dct_headers, timeout=timeout)
      elif method.upper() == "DELETE":
        response = self.requests.delete(url=url, headers=dct_headers, timeout=timeout)
      else:
        raise ValueError("Method {} not implemented in Driver API".format(method))
      if response.status_code in [200, 201]:
        if response.text:
          return response.json()
    except Exception as e:
      raise e

  def __get_url_and_headers(self, headers: dict, path: str):
    """Get the URL and headers for the request.

    Parameters
    ----------
    headers : dict
        The headers to use for the request.
    path : str
        The path to append to the base URL.

    Returns
    -------
    tuple
        The URL and headers for the request.
    """
    base_url = self.cfg_device_ip
    if not base_url.startswith(('http://', 'https://')):
      base_url = "http://" + base_url  # we assume http if not provided

    url = base_url + path

    if headers is None:
      headers = self._headers if hasattr(self, '_headers') else {}

    if not isinstance(headers, dict):
      raise TypeError("The headers must be provided as a dictionary")

    return url, headers



class TorchscriptMixin:

  def _load_torchscript(self, fn_path, device=None):
    """
    Generic method for loading a torchscript and returning both the model generated
    by it and its config.
    Parameters
    ----------
    fn_path - path of the specified torchscript

    Returns
      (model, config) where model is the loaded torchscript model and config is the
        model json config.
    -------

    """
    extra_files = {'config.txt': ''}
    dct_config = None
    model = self.th.jit.load(
      f=fn_path,
      map_location=self.dev,
      _extra_files=extra_files,
    )
    model.eval()
    self.P("Done loading model on device {}".format(
      self.dev,
    ))
    try:
      dct_config = self.json.loads(extra_files['config.txt'].decode('utf-8'))
    except Exception as exc:
      self.P("Could not load in-model config '{}': {}. In future this will stop the model loading: {}".format(
        fn_path, exc, extra_files
      ), color='r')
    return model, dct_config

  def _prepare_ts_model(self, url=None, fn_model=None, post_process_classes=False, return_config=False, **kwargs):
    """
      Do not call directly, use prepare_model instead.
    """
    self.P("Preparing {} torchscript graph model {}...".format(self.server_name, self.version))
    if url is None:
      url = self.get_url
    if url is None:
      raise ValueError("No URL specified for model")
    if fn_model is None:
      fn_model = self.get_saved_model_fn()
    self.download(url, fn_model)
    fn_path = self.log.get_models_file(fn_model)
    model, config = None, None
    if fn_path is None:
      raise ValueError(f"Could not download model file from {url}. {fn_model} not found.")

    self.P("Loading torchscript model {} ({:.03f} MB) at `{}` using map_location: {} on python v{}...".format(
        fn_model,
        self.os_path.getsize(fn_path) / 1024 / 1024,
        fn_path,
        self.dev,
        self.python_version()
      ),
      color='y'
    )
    model, config = self._load_torchscript(fn_path, device=self.dev)
    if self.cfg_fp16:
      model.half()
    err_keys = ['torch']
    env_versions = {
      'python': self.python_version(),
      'torch': self.th.__version__,
      'torchvision': self.tv.__version__
    }
    self.check_versions(config, fn_path, env_versions, err_keys)
    if post_process_classes:
      config = self._process_config_classes(config)
    # endif post_process_classes
    self.P("  Model config:\n{}".format(self.log.dict_pretty_format(config)))
    return (model, config) if return_config else model




from naeural_core.data_structures import GeneralPayload

class _CountingTemplateMixin(object):

  def __init__(self, **kwargs):
    super(_CountingTemplateMixin, self).__init__(**kwargs)
    return

  def _create_payload(self, nr_objects=None, object_type=None, **kwargs):
    if object_type is None:
      object_type = self.cfg_object_type

    if isinstance(object_type, list):
      if len(object_type) == 0:
        object_type = object_type[0]

    payload = GeneralPayload(
      owner=self,
      _v_nr_objects=nr_objects,
      _v_object_type=object_type,
      plugin_category='counting',
      **kwargs
    )
    return payload


from typing import List
from sklearn.metrics import accuracy_score, confusion_matrix, f1_score, recall_score, precision_score
import numpy as np
import torch as th

class _ClassificationMetricsMixin(object):

  def __init__(self):
    super(_ClassificationMetricsMixin, self).__init__()
    return

  def get_classes_accuracies(self, cm, classes=None, key='dev'):
    n_classes = max(cm.shape)
    if classes is None:
      classes = range(n_classes)

    cls_acc = [cm[i, i] for i in range(n_classes)]
    normalized_accuracies = cls_acc / np.sum(cm, axis=1)
    acc_dict = {f'{key}_acc_{classes[i]}': round(normalized_accuracies[i], 2) for i in range(n_classes)}
    return acc_dict

  def basic_metrics(self, y, y_hat, classes, key):
    acc = accuracy_score(y, y_hat)
    cm_labels = list(range(len(classes)))
    cm = confusion_matrix(y, y_hat, labels=cm_labels)
    f1 = f1_score(y, y_hat, average='macro')
    rec = recall_score(y, y_hat, average='macro')
    prec = precision_score(y, y_hat, average='macro')
    classes_accuracies = self.get_classes_accuracies(cm, classes, key=key)

    return {
      '{}_acc'.format(key): acc,
      '{}_cm'.format(key): cm,
      '{}_f1'.format(key) : f1,
      '{}_rec'.format(key) : rec,
      '{}_prec'.format(key) : prec,
      **classes_accuracies
    }

  def advanced_metrics(self, y, y_hat, idx, dataset_info, classes=None, key='dev'):
    dct_metrics_per_categ = {}
    classes = classes if classes is not None else self.get_class_names(dataset_info=dataset_info)

    try:
      dct_level_categs = dataset_info['categ_to_id_per_lvl'][self._level_analysis]
      idpaths = dataset_info['path_to_idpath'][idx][:, self._level_analysis]
    except:
      return dct_metrics_per_categ

    for categ_name, cidx in dct_level_categs.items():
      crt_idx = np.where(idpaths == cidx)[0]
      if isinstance(y, (np.ndarray, th.Tensor)):
        y_crt = y[crt_idx]
        y_hat_crt = y_hat[crt_idx]
      else:
        y_crt = [_y[crt_idx] for _y in y]
        y_hat_crt = [_y[crt_idx] for _y in y_hat]
      #endif

      dct_metrics_per_categ['{}_{}'.format(key, categ_name)] = self.basic_metrics(
        y=y_crt, y_hat=y_hat_crt,
        classes=classes,
        key=key
      )
    #endfor

    return dct_metrics_per_categ

  def log_metrics(self, dct_metrics: dict,
                  classes: List[str],
                  log_cm: bool = True):

    keys = list(dct_metrics.keys())
    lst_cm_keys = []
    str_log = ""
    for k in keys:
      if not isinstance(dct_metrics[k], np.ndarray):
        str_log += "{}: {:.4f} / ".format(k, dct_metrics[k])
      else:
        lst_cm_keys.append(k)

    self.log.P(str_log)
    if log_cm:
      for k in lst_cm_keys:
        cm = dct_metrics[k]
        self.log.P("Printing confusion matrix key '{}'".format(k))
        self.log.log_confusion_matrix(cm, classes=classes)
      #endfor
    #endif
    return



import importlib
import inspect

from naeural_core import constants as ct

from naeural_core.data_structures import MetadataObject

class _DataSensorMixin(object):
  def __init__(self):
    self._data_helper_name = None
    self._data_helper_params = None
    self._training_data_sent = False
    self.sensor = None
    
    super(_DataSensorMixin, self).__init__()
    return

  def _get_sensor_helper_class(self, module_name):
    full_module_name = ct.SENSORS_LOCATION + '.' + module_name
    module = importlib.import_module(full_module_name)
    classes = inspect.getmembers(module, inspect.isclass)
    _class = None
    for _cls in classes:
      if _cls[0].lower() == module_name.lower().replace('_',''):
        _class = _cls[1]
    return _class


  def _start_sensor(self):
    self._data_helper_name = self.cfg_stream_config_metadata[ct.DATA_HELPER_NAME]
    self._data_helper_params = self.cfg_stream_config_metadata[ct.DATA_HELPER_PARAMS]
    if not isinstance(self._data_helper_params, dict):
      raise ValueError("DATA_HELPER_PARAMS must be a dictionary of params for the Sensor class!")
    
    _class = self._get_sensor_helper_class(self._data_helper_name)
    obj = _class(**self._data_helper_params)
    self.sensor = obj
    self.P("Started sensor '{}':".format(obj.__class__.__name__), color='y')
    for k,v in vars(obj).items():
      self.P("  {} : {}".format(
        k,v if isinstance(v, (int, str)) else type(v)),
        color='y')
    self._sensor_metadata = MetadataObject(
      dataframe_count=self.sensor.get_datastream_len(),
      dataframe_current=None,
    )
    return  
  
  def _get_sensor_data(self):
    _train, _obs = None, None
    if not self._training_data_sent and self.sensor.has_training_data():
      self._train_data = self.sensor.get_train_data()
      _train = self._train_data
      self._training_data_sent = True
    else:
      _obs = self.sensor.get_observation()
    _count = self.sensor.get_position()
    return _obs, _train, _count


import numpy as np
import pickle

class _HistogramMixin(object):
  """
  Mixin for histogram functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_HistogramMixin, self).__init__()
    return

  def show_histogram(self, data, bins=10, caption=None, non_negative_only=True,
                     show_both_ends=False):
    return self.show_text_histogram(data, bins,
                                    caption, non_negative_only,
                                    show_both_ends)

  def show_text_histogram(self, data, bins=10, caption=None, non_negative_only=True,
                          show_both_ends=False):
    """
    displays a text histogram of input 1d array
    """
    hist = True
    data = np.array(data)
    if ('u' in data.dtype.str) or ('i' in data.dtype.str):
      hist = False

    if caption is None:
      caption = ''

    if hist:
      caption += " - Histogram"
    else:
      caption += ' - Bin-count'

    if hist:
      res = np.histogram(data, bins=bins)
      y_inp = res[0]
      x_inp = res[1]
      x_format = '{num:{fill}{width}.3f}'
    else:
      x_format = '{num:{fill}{width}.0f}'
      res = np.bincount(data)
      _min = data.min()
      _max = data.max()
      y_inp = res[_min:]
      x_inp = np.arange(_min, _max + 1)
      if non_negative_only:
        non_neg = (y_inp != 0)
        y_inp = y_inp[non_neg]
        x_inp = x_inp[non_neg]
      if bins < y_inp.shape[0]:
        if not show_both_ends:
          cutoff = bins // 2
          caption += ' (showing both {} ends)'.format(cutoff)
          y_inp = np.concatenate((y_inp[:cutoff], y_inp[-cutoff:]))
          x_inp = np.concatenate((x_inp[:cutoff], x_inp[-cutoff:]))
      else:
        show_both_ends = False

    bins = y_inp.shape[0] if bins > y_inp.shape[0] else bins
    _x = []
    _y = []
    if show_both_ends:
      self.P("{} first ({} bins):".format(caption, bins))
      for i in range(bins):
        _y.append('{num:{fill}{width}.0f}'.format(num=y_inp[i], fill=' ', width=8))
        _x.append(x_format.format(num=x_inp[i], fill=' ', width=8))
      self.P("    Y: " + ''.join([y for y in _y]))
      self.P("    X: " + ''.join([x for x in _x]))
      _x = []
      _y = []
      self.P("{} last ({} bins):".format(caption, bins))
      for i in range(x_inp.shape[0] - bins, x_inp.shape[0], 1):
        _y.append('{num:{fill}{width}.0f}'.format(num=y_inp[i], fill=' ', width=8))
        _x.append(x_format.format(num=x_inp[i], fill=' ', width=8))
      self.P("    Y: " + ''.join([y for y in _y]))
      self.P("    X: " + ''.join([x for x in _x]))
    else:
      self.P("{} ({} bins):".format(caption, bins))
      for i in range(bins):
        _y.append('{num:{fill}{width}.0f}'.format(num=y_inp[i], fill=' ', width=8))
        _x.append(x_format.format(num=x_inp[i], fill=' ', width=8))
      self.P("    Y: " + ''.join([y for y in _y]))
      self.P("    X: " + ''.join([x for x in _x]))
    return res

  @staticmethod
  def plot_histogram(distributions, colors=None, legends=None, labels=None,
                     figsize=None, dpi=None, bins=None,
                     logscale=False, xticks=None, rotation_xticks=None, title=None,
                     xlabel=None, ylabel=None, save_img_path=None,
                     close_fig=False, save_fig_pickle=False, xlim=None, ylim=None,
                     vline=None):
    """
    Plots a distribution or multiple distributions (list of ndarrays or lists)
    - distributions: should be a list, where each element in the list is a distribution
                     that will be plotted
    - colors: list which specifies the colors (as string) for each distribution
    """
    # import seaborn as sns
    import matplotlib.pyplot as plt
    plt.style.use('ggplot')
    assert type(distributions) == list, '`distributions` must be a list of lists or list of ndarrays'
    assert type(distributions[0]) in [list, np.ndarray], '`distributions` must be a list of lists or list of ndarrays'

    if legends:
      assert type(legends) == list
      assert len(distributions) == len(legends)

    if colors:
      assert type(colors) == list
      if len(distributions) > len(colors):
        colors += ['blue'] * (len(distributions) - len(colors))
    else:
      colors = ['blue'] * len(distributions)

    n_distrib = len(distributions)
    if figsize and dpi:
      figsize = (figsize[0] / dpi, figsize[1] / dpi)
    elif figsize is None:
      figsize = (13, 8)

    _w = figsize[0]
    _h = figsize[1] * n_distrib
    fig, axs = plt.subplots(n_distrib, 1, figsize=(_w, _h), dpi=dpi)

    if type(axs) not in [list, np.ndarray]:
      axs = [axs]

    for i, distribution in enumerate(distributions):
      # plot = sns.distplot(distribution, hist=True, kde=False,
      #                     bins=bins, color=colors[i], hist_kws={'edgecolor': 'black'})
      label = labels[i] if labels is not None else ''
      axs[i].hist(distribution, color=colors[i], bins=bins)
      axs[i].set_title(label, fontsize=20)
      if xticks is not None:
        axs[i].set_xticks(xticks[i])
      if xlim is not None:
        axs[i].set_xlim(xlim)
      if ylim is not None:
        axs[i].set_ylim(ylim)
      if logscale:
        axs[i].set_yscale('log')
      # if rotation_xticks is not None:
      #   plt.xticks(rotation=rotation_xticks)
      if xlabel is not None:
        axs[i].set(xlabel=xlabel)
      if ylabel is not None:
        axs[i].set(ylabel=ylabel)

      if vline is not None:
        # __min = min(distribution)
        # __max = max(distribution)
        axs[i].axvline(vline, color='k', linestyle='--', linewidth=2)

    if title is not None:  plt.set_title(title)

    fig.tight_layout()
    plt.subplots_adjust(hspace=0.15)

    from .matplotlib_mixin import _MatplotlibMixin
    _MatplotlibMixin.add_copyright_to_plot(plt)

    if save_img_path is not None:
      if not save_img_path.endswith('.png'):
        save_img_path += '.png'
      fig.savefig(save_img_path)
      if save_fig_pickle:
        with open(save_img_path + '.pickle', 'wb') as handle:
          pickle.dump(fig, handle, protocol=pickle.HIGHEST_PROTOCOL)

    if close_fig: plt.close()
    return





"""BK-tree data structure to allow fast querying of "close" matches.
For example usage, see README.rst.
This code is licensed under a permissive MIT license -- see LICENSE.txt.
The pybktree project lives on GitHub here:
https://github.com/benhoyt/pybktree
"""

from collections import deque
from operator import itemgetter
from naeural_core.utils.distance_functions import levenshtein_distance, hamming_distance

__version__ = '1.1'

_getitem0 = itemgetter(0)





class BKTree(object):
    """BK-tree data structure that allows fast querying of matches that are
    "close" given a function to calculate a distance metric (e.g., Hamming
    distance or Levenshtein distance).
    Each node in the tree (including the root node) is a two-tuple of
    (item, children_dict), where children_dict is a dict whose keys are
    non-negative distances of the child to the current item and whose values
    are nodes.
    """
    def __init__(self, distance_func, items=[]):
        """Initialize a BKTree instance with given distance function
        (which takes two items as parameters and returns a non-negative
        distance integer). "items" is an optional list of items to add
        on initialization.
        >>> tree = BKTree(hamming_distance)
        >>> list(tree)
        []
        >>> tree.distance_func is hamming_distance
        True
        >>> tree = BKTree(hamming_distance, [])
        >>> list(tree)
        []
        >>> tree = BKTree(hamming_distance, [0, 4, 5])
        >>> sorted(tree)
        [0, 4, 5]
        """
        self.distance_func = distance_func
        self.tree = None

        _add = self.add
        for item in items:
            _add(item)

    def add(self, item):
        """Add given item to this tree.
        >>> tree = BKTree(hamming_distance)
        >>> list(tree)
        []
        >>> tree.add(4)
        >>> sorted(tree)
        [4]
        >>> tree.add(15)
        >>> sorted(tree)
        [4, 15]
        """
        node = self.tree
        if node is None:
            self.tree = (item, {})
            return

        # Slight speed optimization -- avoid lookups inside the loop
        _distance_func = self.distance_func

        while True:
            parent, children = node
            distance = _distance_func(item, parent)
            node = children.get(distance)
            if node is None:
                children[distance] = (item, {})
                break

    def find(self, item, n):
        """Find items in this tree whose distance is less than or equal to n
        from given item, and return list of (distance, item) tuples ordered by
        distance.
        >>> tree = BKTree(hamming_distance)
        >>> tree.find(13, 1)
        []
        >>> tree.add(0)
        >>> tree.find(1, 1)
        [(1, 0)]
        >>> for item in [0, 4, 5, 14, 15]:
        ...     tree.add(item)
        >>> sorted(tree)
        [0, 0, 4, 5, 14, 15]
        >>> sorted(tree.find(13, 1))
        [(1, 5), (1, 15)]
        >>> sorted(tree.find(13, 2))
        [(1, 5), (1, 15), (2, 4), (2, 14)]
        >>> sorted(tree.find(0, 1000)) == [(hamming_distance(x, 0), x) for x in tree]
        True
        """
        if self.tree is None:
            return []

        candidates = deque([self.tree])
        found = []

        # Slight speed optimization -- avoid lookups inside the loop
        _candidates_popleft = candidates.popleft
        _candidates_extend = candidates.extend
        _found_append = found.append
        _distance_func = self.distance_func

        while candidates:
            candidate, children = _candidates_popleft()
            distance = _distance_func(candidate, item)
            if distance <= n:
                _found_append((distance, candidate))

            if children:
                lower = distance - n
                upper = distance + n
                _candidates_extend(c for d, c in children.items() if lower <= d <= upper)

        found.sort(key=_getitem0)
        return found

    def __iter__(self):
        """Return iterator over all items in this tree; items are yielded in
        arbitrary order.
        >>> tree = BKTree(hamming_distance)
        >>> list(tree)
        []
        >>> tree = BKTree(hamming_distance, [1, 2, 3, 4, 5])
        >>> sorted(tree)
        [1, 2, 3, 4, 5]
        """
        if self.tree is None:
            return

        candidates = deque([self.tree])

        # Slight speed optimization -- avoid lookups inside the loop
        _candidates_popleft = candidates.popleft
        _candidates_extend = candidates.extend

        while candidates:
            candidate, children = _candidates_popleft()
            yield candidate
            _candidates_extend(children.values())

    def __repr__(self):
        """Return a string representation of this BK-tree with a little bit of info.
        >>> BKTree(hamming_distance)
        <BKTree using hamming_distance with no top-level nodes>
        >>> BKTree(hamming_distance, [0, 4, 8, 14, 15])
        <BKTree using hamming_distance with 3 top-level nodes>
        """
        return '<{} using {} with {} top-level nodes>'.format(
            self.__class__.__name__,
            self.distance_func.__name__,
            len(self.tree[1]) if self.tree is not None else 'no',
        )


if __name__ == '__main__':
    bk = BKTree(distance_func=levenshtein_distance)